# coding: utf-8
"""
    GssMixture.py

    Copyright (c) 2020, SAXS Team, KEK-PF
"""
import numpy as np
import scipy.stats as stats
from scipy.optimize import fsolve
from sklearn.cluster import KMeans
import molass_legacy.KekLib.DebugPlot as plt
from .GaussianMixture import gaussian_pdf

USE_MODE_AS_MU_INIT = False
USE_MODERATE_SIGMAS = False
MODERATE_SIGMAS_RATIO = 1.5

class GssMixture:
    def __init__(self, K, max_iter=100, anim_data=False):
        self.K = K
        self.max_iter = max_iter
        self.anim_data = anim_data

    def fit(self, X, bins=None):
        assert X.shape[1] == 1
        self.N = X.shape[0]
        self.X = X
        self.X_  = X.flatten()
        self.bins = bins
        self.initilize()

        for step in range(self.max_iter):
            self._e_step()
            self._m_step(step)

        print('result pi=', self.pi)
        print('result mu=', self.mu)
        print('result sigma=', self.sigma)

    def initilize(self):
        # self._initialise_parameters(self.X)
        self.guess_initial_params()

    def calculate_mean_covariance(self, X, prediction):
        """Calculate means and covariance of different
            clusters from k-means prediction

        Parameters:
        ------------
        prediction: cluster labels from k-means
        X: N*d numpy array data points 

        Returns:
        -------------
        intial_means: for E-step of EM algorithm
        intial_cov: for E-step of EM algorithm

        """
        d = X.shape[1]
        labels = np.unique(prediction)
        initial_means = np.zeros((self.K, d))
        initial_cov = np.zeros((self.K, d, d))
        initial_pi = np.zeros(self.K)
        
        counter=0
        for label in labels:
            ids = np.where(prediction == label) # returns indices
            initial_pi[counter] = len(ids[0]) / X.shape[0]
            initial_means[counter,:] = np.mean(X[ids], axis = 0)
            de_meaned = X[ids] - initial_means[counter,:]
            Nk = X[ids].shape[0] # number of data points in current gaussian
            initial_cov[counter,:, :] = np.dot(initial_pi[counter] * de_meaned.T, de_meaned) / Nk
            counter+=1

        # assert np.sum(initial_pi) == 1
        assert abs(np.sum(initial_pi) - 1) < 1e-10
        return (initial_means, initial_cov, initial_pi)

    def _initialise_parameters(self, X):
        """Implement k-means to find starting
            parameter values.
            https://datascience.stackexchange.com/questions/11487/how-do-i-obtain-the-weight-and-variance-of-a-k-means-cluster

        Parameters:
        ------------
        X: numpy array of data points

        Returns:
        ----------
        tuple containing initial means and covariance
        _initial_means: numpy array: (C*d)
        _initial_cov: numpy array: (C,d*d)

        """
        n_clusters = self.K
        kmeans = KMeans(n_clusters= n_clusters, init="k-means++", max_iter=500, algorithm = 'auto')
        fitted = kmeans.fit(X)
        prediction = kmeans.predict(X)
        _initial_means, _initial_cov, _initial_pi = self.calculate_mean_covariance(X, prediction)

        self.pi = _initial_pi
        self.mu = _initial_means.flatten()
        self.sigma = np.sqrt(_initial_cov.flatten())
        self.tau = np.zeros(self.K)
        print('init pi=', self.pi)
        print('init mu=', self.mu)
        print('init sigma=', self.sigma)
        print('init tau=', self.tau)

    def guess_initial_params(self):
        X = self.X
        num_entire_points = X.shape[0]
        kmeans = KMeans(n_clusters= self.K)
        kmeans.fit(X)
        predicted = kmeans.predict(X)
        labels = np.unique(predicted)
        init_pi = []
        init_mu = []
        init_sigma = []

        for label in labels:
            sy = X[predicted == label]

            M1 = np.mean(sy)
            if USE_MODE_AS_MU_INIT:
                # seems risky
                m, c = stats.mode(sy)
                mu = m[0,0]
            else:
                mu = M1

            init_mu.append(mu)
            num_points = sy.shape[0]
            pi = num_points/num_entire_points
            init_pi.append(pi)
            # M2 = pi*np.sum((sy - mu)**2)/num_points     # why pi* in Foley's implementation?
            M2 = np.sum((sy - M1)**2)/num_points
            sigma = np.sqrt(M2)
            # print([label], m, c, M2, np.sqrt(M2), sigma)
            init_sigma.append(sigma)

        self.pi = np.array(init_pi)
        self.mu = np.array(init_mu)
        self.sigma = np.array(init_sigma)
        print('init pi=', self.pi)
        print('init mu=', self.mu)
        print('init sigma=', self.sigma)

        # assert np.sum(initial_pi) == 1
        assert abs(np.sum(self.pi) - 1) < 1e-10

        if False:
            plt.push()
            fig = plt.figure()
            ax = fig.gca()
            ax.hist(X.flatten(), bins=self.bins)
            ymin, ymax = ax.get_ylim()
            ax.set_ylim(ymin, ymax)
            for mu in self.mu:
                ax.plot([mu, mu], [ymin, ymax], ':', color='red')
            plt.show()
            plt.pop()

        if self.anim_data:
            array_shape = (self.max_iter+1, self.K)
            self.pi_array = np.zeros(array_shape)
            self.mu_array = np.zeros(array_shape)
            self.sigma_array = np.zeros(array_shape)
            self.set_anim_params(0)

    def set_anim_params(self, n):
        self.pi_array[n,:] = self.pi
        self.mu_array[n,:] = self.mu
        self.sigma_array[n,:] = self.sigma

    def _e_step(self):
        self.gamma = np.zeros((self.N, self.K))

        for k in range(self.K):
            # Posterior Distribution using Bayes Rule
            self.gamma[:,k] = self.pi[k] * gaussian_pdf(self.X_, self.mu[k], self.sigma[k])

        # normalize across columns to make a valid probability
        # gamma_norm = np.sum(self.gamma, axis=1)[:,np.newaxis]
        gamma_norm = np.sum(self.gamma, axis=1)

        if False:
            print('pi=', self.pi)
            print('gamma_norm=', gamma_norm)
            plt.push()
            fig = plt.figure()
            ax = fig.gca()
            ax.plot(gamma_norm.flatten())
            plt.show()
            plt.pop()

        positive = gamma_norm > 0
        for k in range(self.K):
            self.gamma[positive,k] /= gamma_norm[positive]      # normalize only where positive

    def _m_step(self, step):

        # responsibilities for each gaussian
        self.pi = np.mean(self.gamma, axis=0)

        W_ = np.sum(self.gamma, axis=0)[:,np.newaxis].T
        # print('GT.shape=', GT.shape)
        # print('GT[:,0:5]=', GT[:,0:5])
        # print('W_=', W_)

        """
        2.1.1 Central Moments and their Distribution Functions
        https://www.iue.tuwien.ac.at/phd/puchner/node17.html#SECTION00811000000000000000
        2.1.2 Probability Weighted Moments and their Distribution Functions
        https://www.iue.tuwien.ac.at/phd/puchner/node18.html

        Central moment - Wikipedia
        https://en.wikipedia.org/wiki/Central_moment
        """

        debug = False

        # M1 = (np.dot(self.gamma.T, self.X) / W_.T).T
        M1 = np.sum(self.gamma * self.X, axis=0) / W_
        M2 = np.sum(self.gamma * (self.X-M1)**2, axis=0) / W_

        if debug:
            print('self.gamma[:5,:]=', self.gamma[:5,:])
            print('M1=', M1)
            print('M2=', M2)

        params = []
        for k in range(self.K):
            params.append(self.solve_params(k, M1[0,k], M2[0,k]))

        params_ = np.array(params)
        if debug:
            print('params_=', params_)
        self.mu = params_[:,0]
        if USE_MODERATE_SIGMAS:
            self.sigma = self.moderate_sigmas(params_[:,1])
        else:
            self.sigma = params_[:,1]

        if self.anim_data:
            self.set_anim_params(step+1)

    def moderate_sigmas(self, sigmas):
        if self.K > 1:
            L = list(range(self.K))
            k = np.argmax(sigmas)
            del L[k]
            sigmas_ = sigmas.copy()
            sigmas_[k] = min(sigmas[k], np.average(sigmas[L]) * MODERATE_SIGMAS_RATIO)
            return sigmas_
        else:
            return sigmas

    def solve_params(self, k, m1, m2):
        last_mu = self.mu[k]
        last_sigma = self.sigma[k]

        mu = m1
        sigma = np.sqrt(m2)

        debug = False

        if debug:
            print([k], (last_mu, '->', mu))
            print([k], (last_sigma, '->', sigma))

        return mu, sigma



# coding: utf-8
"""
    SecTheory.RealisticSolver.py

    Copyright (c) 2022, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.optimize import minimize, basinhopping
from scipy.interpolate import UnivariateSpline
import molass_legacy.KekLib.DebugPlot as plt
from .SecPDF import compute_standard_wCD

class RealisticCfDomain:
    def __init__(self, x, y, cf):
        self.x = x
        self.y = y
        self.cf = cf
        self.area = np.sum(y)
        y_ = y/self.area
        self.N = N = 1024
        w, C, D = compute_standard_wCD(N)
        self.w = w
        self.C = C
        self.D = D
        cft = []
        for w_ in w:
            cft.append(np.sum(np.exp(1j*w_*x)*y_))
        self.cft = np.array(cft)

    def fit(self, init_params):

        def obj_func(p):
            return self.cf.objective_func(self.w, p, self.cft)

        opt_result = basinhopping(obj_func, init_params)
        return opt_result

    def get_spline(self, opt_params):
        N = self.N
        cft = self.cf(self.w[N//2:], opt_params)
        cft = np.concatenate([cft[::-1].conj(), cft])
        pdfFFT = np.max([np.zeros(N), (self.C*np.fft.fft(self.D*cft)).real], axis=0)
        spline = UnivariateSpline(np.arange(N), pdfFFT*self.area, s=0)
        return spline

    def get_component_splines(self, opt_params):
        N = self.N
        half_w = self.w[N//2:]
        zeros = np.zeros(N)
        x = np.arange(N)
        model_cf = self.cf.model_cf

        rg_params, rp_params = self.cf.split_params(opt_params)
        ret_splines = []
        for rg, weight in rg_params:
            cft = model_cf(half_w, rg, *rp_params)
            cft = np.concatenate([cft[::-1].conj(), cft])
            pdfFFT = np.max([zeros, (self.C*np.fft.fft(self.D*cft)).real], axis=0)
            spline = UnivariateSpline(x, pdfFFT*weight*self.area, s=0)
            ret_splines.append(spline)
        return ret_splines



"""
    Selective.DataFilter.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import logging
import numpy as np
from molass_legacy.Models.ElutionCurveModels import EGH, EGHA
from Selective.LrfSource import LrfSource
from molass_legacy.Models.ModelUtils import compute_cy_list, compute_area_props

class DataFilter(LrfSource):
    def __init__(self, lrf_src):
        self.logger = logging.getLogger(__name__)
        self.apply_filter(lrf_src)
    
    def apply_filter(self, lrf_src):
        self.corrected_sd = lrf_src.corrected_sd
        self.uv_x = lrf_src.uv_x
        self.uv_y = lrf_src.uv_y
        self.xr_x = lrf_src.xr_x
        self.xr_y = lrf_src.xr_y
        self.baselines = lrf_src.baselines
        uv_peaks = lrf_src.uv_peaks
        xr_peaks = lrf_src.xr_peaks
        if xr_peaks.shape[1] == 4:
            self.model = EGH()
        else:
            self.model = EGHA()
        indeces = self.make_filter_indeces(uv_peaks, xr_peaks)
        self.uv_peaks = uv_peaks[indeces]
        self.xr_peaks = xr_peaks[indeces]
        self.rg_info = None
        self.egh_moments_list = None       
        # self.draw()
        self._compute_rgs()

    def make_filter_indeces(self, uv_peaks, xr_peaks):
        # AhRR: remove 0 for the reason of minor peaks near a major peak top
        num_peaks = len(xr_peaks)
        cy_list = compute_cy_list(self.model, self.xr_x, xr_peaks)
        props = compute_area_props(cy_list)
        full_indeces = np.arange(num_peaks)
        remove_indeces = []
        topxv = xr_peaks[:,1]
        max_sigma = np.max(xr_peaks[:,2])
        print("props=", props, "topxv=", topxv, "max_sigma=", max_sigma)
        for k, p in enumerate(props):
            if p < 0.05:
                if k < num_peaks-1:
                    dx = topxv[k+1] - topxv[k]
                    if dx/max_sigma < 1:
                        remove_indeces.append(k)
                if k > 0:
                    dx = topxv[k] - topxv[k-1]
                    if dx/max_sigma < 1:
                        remove_indeces.append(k)
        if len(remove_indeces) > 0:
            indeces = np.setdiff1d(full_indeces, remove_indeces)
            self.logger.info("make_filter_indeces: filtering by %s as in AhRR with a centered minor peak", indeces)
        else:
            indeces = full_indeces
        return indeces

    def get_filtered_src(self):
        return self


"""
    Selective.AdvancedAdapter.py

    Copyright (c) 2024-2025, SAXS Team, KEK-PF
"""

def adapted_decompose_elution_simply(self):
    from molass_legacy.KekLib.BasicUtils import Struct
    from importlib import reload
    import molass_legacy.CFSD.SimpleDecompose
    reload(molass_legacy.CFSD.SimpleDecompose)
    from molass_legacy.CFSD.SimpleDecompose import decompose_elution_simply
    x = self.corbase_info.x
    y = self.corbase_info.y
    dialog = self.dialog
    min_frame = dialog.frames[dialog.min_index]
    uv_y = min_frame.uv_y
    num_peaks = len(min_frame.opt_recs)
    traditional_info = Struct(num_peaks=num_peaks, mapper=min_frame.mapper)
    decomp_result = decompose_elution_simply(x, y, uv_y, self.model, traditional_info)
    return decomp_result


"""
    Estimators.BaseEstimator.py

    Copyright (c) 2022-2025, SAXS Team, KEK-PF
"""
import logging
from molass_legacy._MOLASS.SerialSettings import set_setting
from molass_legacy.SecTheory.T0UpperBound import estimate_t0upper_bound

class BaseEstimator:
    def __init__(self, editor):
        self.logger = logging.getLogger(self.__class__.__name__)
        self.editor = editor
        self.sd = editor.sd
        self.corrected_sd = editor.corrected_sd
        self.nc = editor.get_n_components()     # this includes baseline
        self.ecurves = editor.ecurves
        self.t0_upper_bound = None

    def get_t0_upper_bound(self):
        if self.t0_upper_bound is None:
            ecurve = self.sd.get_xray_curve()
            t0_upper_bound = estimate_t0upper_bound(ecurve)
            self.logger.info("t0_upper_bound has been estimated to be %g", t0_upper_bound)
            set_setting("t0_upper_bound", t0_upper_bound)
            self.t0_upper_bound = t0_upper_bound
        return self.t0_upper_bound



# coding: utf-8
"""
    RegalsDemo.py

    Copyright (c) 2021, SAXS Team, KEK-PF
"""
import numpy as np
from .regals import component, concentration_class, profile_class, mixture, regals
from SvdDenoise import get_denoised_data
import molass_legacy.KekLib.DebugPlot as plt

def demo(sp, sd):
    pdata, popts = sp.get_lrf_args(sd)

    # print(pdata.paired_ranges)

    D, E, q, ecurve = sd.get_xr_data_separate_ly()
    print(D.shape)

    I = D
    sigma = E

    x = ecurve.x
    y = ecurve.y

    C_list = []
    decomp_info = pdata.decomp_info
    for rec in decomp_info.opt_recs:
        f = rec[1]
        C_list.append(f(x))

    C_ = np.array(C_list)
    D_ = get_denoised_data(D, rank=2)
    P_ = D_ @ np.linalg.pinv(C_)

    C1 = component(concentration_class('smooth', x, xmin=0, xmax=120), profile_class('simple', q))
    C2 = component(concentration_class('smooth', x, xmin=70, xmax=280), profile_class('simple', q))
    M = mixture([C1, C2])
    M.lambda_concentration = M.estimate_concentration_lambda(E, np.array([np.inf, np.inf]))
    R = regals(I, sigma)

    def stop_fun(num_iter, params):
        return [num_iter >= 100, 'max_iter']

    def update_fun(num_iter, mix, params, resid):
        print('%2d, x2 = %f, delta_profile = %s'
                % (num_iter, params['x2'], np.array2string(params['delta_profile'], precision=3)))

    [M1,params,resid,exit_cond] = R.run(M, stop_fun, update_fun)
    [I1, sigma1] = M1.extract_profile(I, sigma, 0)
    [I2, sigma2] = M1.extract_profile(I, sigma, 1)

    fig, axes = plt.subplots(nrows=2, ncols=2, figsize=(16,11))
    fig.suptitle("Comparison between _MOLASS and REGALS on Elution Decomposition", fontsize=20)
    ax1 = axes[0,0]
    ax2 = axes[0,1]
    ax3 = axes[1,0]
    ax4 = axes[1,1]
    ax1.set_title("Decomposition from molass_legacy._MOLASS", fontsize=16)
    ax2.set_title("Decomposition from REGALS", fontsize=16)
    ax3.set_title("Profiles from molass_legacy._MOLASS", fontsize=16)
    ax4.set_title("Profiles from REGALS", fontsize=16)

    for ax in [ax3, ax4]:
        ax.set_yscale('log')

    for ax in [ax1, ax2]:
        ax.plot(x, y, color='orange')

    for c in C_list:
        ax1.plot(x, c, ':')

    ax2t = ax2.twinx()
    ax2t.grid(False)
    for k, c in enumerate(M1.concentrations.T):
        ax2t.plot(x, c, ':')


    for p in P_.T:
        ax3.plot(q, p)

    for I_ in [I1, I2]:
        ax4.plot(q, I_)

    fig.tight_layout()
    fig.subplots_adjust(top=0.9)
    plt.show()



"""
    LrfRgUtils.py

    Copyright (c) 2023, SAXS Team, KEK-PF
"""
import numpy as np
from molass_legacy.GuinierAnalyzer.SimpleGuinier import SimpleGuinier

def compute_rg_from_qvDEP(qv, D, E, P, return_sg=False):
    aq = P.T[0]
    Dinv = np.linalg.pinv(D)
    W = np.dot(Dinv, P)
    Pe = np.sqrt(np.dot(E**2, W**2))
    data = np.array([qv, aq, Pe[:,0]]).T

    sg = SimpleGuinier(data)
    if return_sg:
        return sg
    else:
        return sg.Rg



#   Copyright 2024 The PyMC Developers
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Functions for MCMC sampling."""

import contextlib
import logging
import pickle
import sys
import time
import warnings

from collections.abc import Callable, Iterator, Mapping, Sequence
from typing import (
    Any,
    Literal,
    TypeAlias,
    overload,
)

import numpy as np
import pytensor.gradient as tg

from arviz import InferenceData, dict_to_dataset
from arviz.data.base import make_attrs
from pytensor.graph.basic import Variable
from rich.console import Console
from rich.progress import BarColumn, TextColumn, TimeElapsedColumn, TimeRemainingColumn
from rich.theme import Theme
from threadpoolctl import threadpool_limits
from typing_extensions import Protocol

import pymc as pm

from pymc.backends import RunType, TraceOrBackend, init_traces
from pymc.backends.arviz import (
    coords_and_dims_for_inferencedata,
    find_constants,
    find_observations,
)
from pymc.backends.base import IBaseTrace, MultiTrace, _choose_chains
from pymc.blocking import DictToArrayBijection
from pymc.exceptions import SamplingError
from pymc.initial_point import PointType, StartDict, make_initial_point_fns_per_chain
from pymc.model import Model, modelcontext
from pymc.sampling.parallel import Draw, _cpu_count
from pymc.sampling.population import _sample_population
from pymc.stats.convergence import (
    log_warning_stats,
    log_warnings,
    run_convergence_checks,
)
from pymc.step_methods import NUTS, CompoundStep
from pymc.step_methods.arraystep import BlockedStep, PopulationArrayStepShared
from pymc.step_methods.hmc import quadpotential
from pymc.util import (
    CustomProgress,
    RandomSeed,
    RandomState,
    _get_seeds_per_chain,
    default_progress_theme,
    drop_warning_stat,
    get_untransformed_name,
    is_transformed_name,
)
from pymc.vartypes import discrete_types

sys.setrecursionlimit(10000)

__all__ = [
    "sample",
    "init_nuts",
]

Step: TypeAlias = BlockedStep | CompoundStep


class SamplingIteratorCallback(Protocol):
    """Signature of the callable that may be passed to `pm.sample(callable=...)`."""

    def __call__(self, trace: IBaseTrace, draw: Draw):
        pass


_log = logging.getLogger(__name__)


def instantiate_steppers(
    model: Model,
    steps: list[Step],
    selected_steps: Mapping[type[BlockedStep], list[Any]],
    step_kwargs: dict[str, dict] | None = None,
) -> Step | list[Step]:
    """Instantiate steppers assigned to the model variables.

    This function is intended to be called automatically from ``sample()``, but
    may be called manually.

    Parameters
    ----------
    model : Model object
        A fully-specified model object.
    steps : list, array_like of shape (selected_steps, )
        A list of zero or more step function instances that have been assigned to some subset of
        the model's parameters.
    selected_steps : dict
        A dictionary that maps a step method class to a list of zero or more model variables.
    step_kwargs : dict, default=None
        Parameters for the samplers. Keys are the lower case names of
        the step method, values a dict of arguments. Defaults to None.

    Returns
    -------
    methods : list or step
        List of step methods associated with the model's variables, or step method
        if there is only one.
    """
    if step_kwargs is None:
        step_kwargs = {}

    used_keys = set()
    for step_class, vars in selected_steps.items():
        if vars:
            name = getattr(step_class, "name")
            args = step_kwargs.get(name, {})
            used_keys.add(name)
            step = step_class(vars=vars, model=model, **args)
            steps.append(step)

    unused_args = set(step_kwargs).difference(used_keys)
    if unused_args:
        s = "s" if len(unused_args) > 1 else ""
        example_arg = sorted(unused_args)[0]
        example_step = (list(selected_steps.keys()) or pm.STEP_METHODS)[0]
        example_step_name = getattr(example_step, "name")
        raise ValueError(
            f"Invalid key{s} found in step_kwargs: {unused_args}. "
            "Keys must be step names and values valid kwargs for that stepper. "
            f'Did you mean {{"{example_step_name}": {{"{example_arg}": ...}}}}?'
        )

    if len(steps) == 1:
        return steps[0]

    return steps


def assign_step_methods(
    model: Model,
    step: Step | Sequence[Step] | None = None,
    methods: Sequence[type[BlockedStep]] | None = None,
    step_kwargs: dict[str, Any] | None = None,
) -> Step | list[Step]:
    """Assign model variables to appropriate step methods.

    Passing a specified model will auto-assign its constituent stochastic
    variables to step methods based on the characteristics of the variables.
    This function is intended to be called automatically from ``sample()``, but
    may be called manually. Each step method passed should have a
    ``competence()`` method that returns an ordinal competence value
    corresponding to the variable passed to it. This value quantifies the
    appropriateness of the step method for sampling the variable.

    Parameters
    ----------
    model : Model object
        A fully-specified model object.
    step : step function or iterable of step functions, optional
        One or more step functions that have been assigned to some subset of
        the model's parameters. Defaults to ``None`` (no assigned variables).
    methods : iterable of step method classes, optional
        The set of step methods from which the function may choose. Defaults
        to the main step methods provided by PyMC.
    step_kwargs : dict, optional
        Parameters for the samplers. Keys are the lower case names of
        the step method, values a dict of arguments.

    Returns
    -------
    methods : list
        List of step methods associated with the model's variables.
    """
    steps: list[Step] = []
    assigned_vars: set[Variable] = set()

    if step is not None:
        if isinstance(step, BlockedStep | CompoundStep):
            steps.append(step)
        else:
            steps.extend(step)
        for step in steps:
            for var in step.vars:
                if var not in model.value_vars:
                    raise ValueError(
                        f"{var} assigned to {step} sampler is not a value variable in the model. "
                        "You can use `util.get_value_vars_from_user_vars` to parse user provided variables."
                    )
            assigned_vars = assigned_vars.union(set(step.vars))

    # Use competence classmethods to select step methods for remaining
    # variables
    methods_list: list[type[BlockedStep]] = list(methods or pm.STEP_METHODS)
    selected_steps: dict[type[BlockedStep], list] = {}
    model_logp = model.logp()

    for var in model.value_vars:
        if var not in assigned_vars:
            # determine if a gradient can be computed
            has_gradient = getattr(var, "dtype") not in discrete_types
            if has_gradient:
                import pytensor
                try:
                    tg.grad(model_logp, var)  # type: ignore
                except (NotImplementedError, tg.NullTypeGradError, pytensor.gradient.DisconnectedInputError):
                    has_gradient = False

            # select the best method
            rv_var = model.values_to_rvs[var]
            selected = max(
                methods_list,
                key=lambda method, var=rv_var, has_gradient=has_gradient: method._competence(  # type: ignore
                    var, has_gradient
                ),
            )
            selected_steps.setdefault(selected, []).append(var)

    return instantiate_steppers(model, steps, selected_steps, step_kwargs)


def _print_step_hierarchy(s: Step, level: int = 0) -> None:
    if isinstance(s, CompoundStep):
        _log.info(">" * level + "CompoundStep")
        for i in s.methods:
            _print_step_hierarchy(i, level + 1)
    else:
        varnames = ", ".join(
            [
                get_untransformed_name(v.name) if is_transformed_name(v.name) else v.name
                for v in s.vars
            ]
        )
        _log.info(">" * level + f"{s.__class__.__name__}: [{varnames}]")


def all_continuous(vars):
    """Check that vars not include discrete variables"""
    if any([(var.dtype in discrete_types) for var in vars]):
        return False
    else:
        return True


def _sample_external_nuts(
    sampler: Literal["nutpie", "numpyro", "blackjax"],
    draws: int,
    tune: int,
    chains: int,
    target_accept: float,
    random_seed: RandomState | None,
    initvals: StartDict | Sequence[StartDict | None] | None,
    model: Model,
    var_names: Sequence[str] | None,
    progressbar: bool,
    idata_kwargs: dict | None,
    compute_convergence_checks: bool,
    nuts_sampler_kwargs: dict | None,
    **kwargs,
):
    if nuts_sampler_kwargs is None:
        nuts_sampler_kwargs = {}

    if sampler == "nutpie":
        try:
            import nutpie
        except ImportError as err:
            raise ImportError(
                "nutpie not found. Install it with conda install -c conda-forge nutpie"
            ) from err

        if initvals is not None:
            warnings.warn(
                "`initvals` are currently not passed to nutpie sampler. "
                "Use `init_mean` kwarg following nutpie specification instead.",
                UserWarning,
            )

        if idata_kwargs is not None:
            warnings.warn(
                "`idata_kwargs` are currently ignored by the nutpie sampler",
                UserWarning,
            )
        if var_names is not None:
            warnings.warn(
                "`var_names` are currently ignored by the nutpie sampler",
                UserWarning,
            )
        compiled_model = nutpie.compile_pymc_model(model)
        t_start = time.time()
        idata = nutpie.sample(
            compiled_model,
            draws=draws,
            tune=tune,
            chains=chains,
            target_accept=target_accept,
            seed=_get_seeds_per_chain(random_seed, 1)[0],
            progress_bar=progressbar,
            **nuts_sampler_kwargs,
        )
        t_sample = time.time() - t_start
        # Temporary work-around. Revert once https://github.com/pymc-devs/nutpie/issues/74 is fixed
        # gather observed and constant data as nutpie.sample() has no access to the PyMC model
        coords, dims = coords_and_dims_for_inferencedata(model)
        constant_data = dict_to_dataset(
            find_constants(model),
            library=pm,
            coords=coords,
            dims=dims,
            default_dims=[],
        )
        observed_data = dict_to_dataset(
            find_observations(model),
            library=pm,
            coords=coords,
            dims=dims,
            default_dims=[],
        )
        attrs = make_attrs(
            {
                "sampling_time": t_sample,
                "tuning_steps": tune,
            },
            library=nutpie,
        )
        for k, v in attrs.items():
            idata.posterior.attrs[k] = v
        idata.add_groups(
            {"constant_data": constant_data, "observed_data": observed_data},
            coords=coords,
            dims=dims,
        )
        return idata

    elif sampler in ("numpyro", "blackjax"):
        import pymc.sampling.jax as pymc_jax

        idata = pymc_jax.sample_jax_nuts(
            draws=draws,
            tune=tune,
            chains=chains,
            target_accept=target_accept,
            random_seed=random_seed,
            initvals=initvals,
            model=model,
            var_names=var_names,
            progressbar=progressbar,
            nuts_sampler=sampler,
            idata_kwargs=idata_kwargs,
            compute_convergence_checks=compute_convergence_checks,
            **nuts_sampler_kwargs,
        )
        return idata

    else:
        raise ValueError(
            f"Sampler {sampler} not found. Choose one of ['nutpie', 'numpyro', 'blackjax', 'pymc']."
        )


@overload
def sample(
    draws: int = 1000,
    *,
    tune: int = 1000,
    chains: int | None = None,
    cores: int | None = None,
    random_seed: RandomState = None,
    progressbar: bool = True,
    progressbar_theme: Theme | None = default_progress_theme,
    step=None,
    var_names: Sequence[str] | None = None,
    nuts_sampler: Literal["pymc", "nutpie", "numpyro", "blackjax"] = "pymc",
    initvals: StartDict | Sequence[StartDict | None] | None = None,
    init: str = "auto",
    jitter_max_retries: int = 10,
    n_init: int = 200_000,
    trace: TraceOrBackend | None = None,
    discard_tuned_samples: bool = True,
    compute_convergence_checks: bool = True,
    keep_warning_stat: bool = False,
    return_inferencedata: Literal[True] = True,
    idata_kwargs: dict[str, Any] | None = None,
    nuts_sampler_kwargs: dict[str, Any] | None = None,
    callback=None,
    mp_ctx=None,
    blas_cores: int | None | Literal["auto"] = "auto",
    **kwargs,
) -> InferenceData: ...


@overload
def sample(
    draws: int = 1000,
    *,
    tune: int = 1000,
    chains: int | None = None,
    cores: int | None = None,
    random_seed: RandomState = None,
    progressbar: bool = True,
    progressbar_theme: Theme | None = default_progress_theme,
    step=None,
    var_names: Sequence[str] | None = None,
    nuts_sampler: Literal["pymc", "nutpie", "numpyro", "blackjax"] = "pymc",
    initvals: StartDict | Sequence[StartDict | None] | None = None,
    init: str = "auto",
    jitter_max_retries: int = 10,
    n_init: int = 200_000,
    trace: TraceOrBackend | None = None,
    discard_tuned_samples: bool = True,
    compute_convergence_checks: bool = True,
    keep_warning_stat: bool = False,
    return_inferencedata: Literal[False],
    idata_kwargs: dict[str, Any] | None = None,
    nuts_sampler_kwargs: dict[str, Any] | None = None,
    callback=None,
    mp_ctx=None,
    model: Model | None = None,
    blas_cores: int | None | Literal["auto"] = "auto",
    **kwargs,
) -> MultiTrace: ...


def sample(
    draws: int = 1000,
    *,
    tune: int = 1000,
    chains: int | None = None,
    cores: int | None = None,
    random_seed: RandomState = None,
    progressbar: bool = True,
    progressbar_theme: Theme | None = default_progress_theme,
    step=None,
    var_names: Sequence[str] | None = None,
    nuts_sampler: Literal["pymc", "nutpie", "numpyro", "blackjax"] = "pymc",
    initvals: StartDict | Sequence[StartDict | None] | None = None,
    init: str = "auto",
    jitter_max_retries: int = 10,
    n_init: int = 200_000,
    trace: TraceOrBackend | None = None,
    discard_tuned_samples: bool = True,
    compute_convergence_checks: bool = True,
    keep_warning_stat: bool = False,
    return_inferencedata: bool = True,
    idata_kwargs: dict[str, Any] | None = None,
    nuts_sampler_kwargs: dict[str, Any] | None = None,
    callback=None,
    mp_ctx=None,
    blas_cores: int | None | Literal["auto"] = "auto",
    model: Model | None = None,
    **kwargs,
) -> InferenceData | MultiTrace:
    r"""Draw samples from the posterior using the given step methods.

    Multiple step methods are supported via compound step methods.

    Parameters
    ----------
    draws : int
        The number of samples to draw. Defaults to 1000. The number of tuned samples are discarded
        by default. See ``discard_tuned_samples``.
    tune : int
        Number of iterations to tune, defaults to 1000. Samplers adjust the step sizes, scalings or
        similar during tuning. Tuning samples will be drawn in addition to the number specified in
        the ``draws`` argument, and will be discarded unless ``discard_tuned_samples`` is set to
        False.
    chains : int
        The number of chains to sample. Running independent chains is important for some
        convergence statistics and can also reveal multiple modes in the posterior. If ``None``,
        then set to either ``cores`` or 2, whichever is larger.
    cores : int
        The number of chains to run in parallel. If ``None``, set to the number of CPUs in the
        system, but at most 4.
    random_seed : int, array-like of int, RandomState or Generator, optional
        Random seed(s) used by the sampling steps. If a list, tuple or array of ints
        is passed, each entry will be used to seed each chain. A ValueError will be
        raised if the length does not match the number of chains.
    progressbar : bool, optional default=True
        Whether or not to display a progress bar in the command line. The bar shows the percentage
        of completion, the sampling speed in samples per second (SPS), and the estimated remaining
        time until completion ("expected time of arrival"; ETA).
        Only applicable to the pymc nuts sampler.
    step : function or iterable of functions
        A step function or collection of functions. If there are variables without step methods,
        step methods for those variables will be assigned automatically. By default the NUTS step
        method will be used, if appropriate to the model.
    var_names : list of str, optional
        Names of variables to be stored in the trace. Defaults to all free variables and deterministics.
    nuts_sampler : str
        Which NUTS implementation to run. One of ["pymc", "nutpie", "blackjax", "numpyro"].
        This requires the chosen sampler to be installed.
        All samplers, except "pymc", require the full model to be continuous.
    blas_cores: int or "auto" or None, default = "auto"
        The total number of threads blas and openmp functions should use during sampling.
        Setting it to "auto" will ensure that the total number of active blas threads is the
        same as the `cores` argument. If set to an integer, the sampler will try to use that total
        number of blas threads. If `blas_cores` is not divisible by `cores`, it might get rounded
        down. If set to None, this will keep the default behavior of whatever blas implementation
        is used at runtime.
    initvals : optional, dict, array of dict
        Dict or list of dicts with initial value strategies to use instead of the defaults from
        `Model.initial_values`. The keys should be names of transformed random variables.
        Initialization methods for NUTS (see ``init`` keyword) can overwrite the default.
    init : str
        Initialization method to use for auto-assigned NUTS samplers. See `pm.init_nuts` for a list
        of all options. This argument is ignored when manually passing the NUTS step method.
        Only applicable to the pymc nuts sampler.
    jitter_max_retries : int
        Maximum number of repeated attempts (per chain) at creating an initial matrix with uniform
        jitter that yields a finite probability. This applies to ``jitter+adapt_diag`` and
        ``jitter+adapt_full`` init methods.
    n_init : int
        Number of iterations of initializer. Only works for 'ADVI' init methods.
    trace : backend, optional
        A backend instance or None.
        If None, the NDArray backend is used.
    discard_tuned_samples : bool
        Whether to discard posterior samples of the tune interval.
    compute_convergence_checks : bool, default=True
        Whether to compute sampler statistics like Gelman-Rubin and ``effective_n``.
    keep_warning_stat : bool
        If ``True`` the "warning" stat emitted by, for example, HMC samplers will be kept
        in the returned ``idata.sample_stats`` group.
        This leads to the ``idata`` not supporting ``.to_netcdf()`` or ``.to_zarr()`` and
        should only be set to ``True`` if you intend to use the "warning" objects right away.
        Defaults to ``False`` such that ``pm.drop_warning_stat`` is applied automatically,
        making the ``InferenceData`` compatible with saving.
    return_inferencedata : bool
        Whether to return the trace as an :class:`arviz:arviz.InferenceData` (True) object or a
        `MultiTrace` (False). Defaults to `True`.
    idata_kwargs : dict, optional
        Keyword arguments for :func:`pymc.to_inference_data`
    nuts_sampler_kwargs : dict, optional
        Keyword arguments for the sampling library that implements nuts.
        Only used when an external sampler is specified via the `nuts_sampler` kwarg.
    callback : function, default=None
        A function which gets called for every sample from the trace of a chain. The function is
        called with the trace and the current draw and will contain all samples for a single trace.
        the ``draw.chain`` argument can be used to determine which of the active chains the sample
        is drawn from.
        Sampling can be interrupted by throwing a ``KeyboardInterrupt`` in the callback.
    mp_ctx : multiprocessing.context.BaseContent
        A multiprocessing context for parallel sampling.
        See multiprocessing documentation for details.
    model : Model (optional if in ``with`` context)
        Model to sample from. The model needs to have free random variables.

    Returns
    -------
    trace : pymc.backends.base.MultiTrace or arviz.InferenceData
        A ``MultiTrace`` or ArviZ ``InferenceData`` object that contains the samples.

    Notes
    -----
    Optional keyword arguments can be passed to ``sample`` to be delivered to the
    ``step_method``\ s used during sampling.

    For example:

       1. ``target_accept`` to NUTS: nuts={'target_accept':0.9}
       2. ``transit_p`` to BinaryGibbsMetropolis: binary_gibbs_metropolis={'transit_p':.7}

    Note that available step names are:

    ``nuts``, ``hmc``, ``metropolis``, ``binary_metropolis``,
    ``binary_gibbs_metropolis``, ``categorical_gibbs_metropolis``,
    ``DEMetropolis``, ``DEMetropolisZ``, ``slice``

    The NUTS step method has several options including:

        * target_accept : float in [0, 1]. The step size is tuned such that we
          approximate this acceptance rate. Higher values like 0.9 or 0.95 often
          work better for problematic posteriors. This argument can be passed directly to sample.
        * max_treedepth : The maximum depth of the trajectory tree
        * step_scale : float, default 0.25
          The initial guess for the step size scaled down by :math:`1/n**(1/4)`,
          where n is the dimensionality of the parameter space

    Alternatively, if you manually declare the ``step_method``\ s, within the ``step``
       kwarg, then you can address the ``step_method`` kwargs directly.
       e.g. for a CompoundStep comprising NUTS and BinaryGibbsMetropolis,
       you could send ::

        step = [
            pm.NUTS([freeRV1, freeRV2], target_accept=0.9),
            pm.BinaryGibbsMetropolis([freeRV3], transit_p=0.7),
        ]

    You can find a full list of arguments in the docstring of the step methods.

    Examples
    --------
    .. code:: ipython

        In [1]: import pymc as pm
           ...: n = 100
           ...: h = 61
           ...: alpha = 2
           ...: beta = 2

        In [2]: with pm.Model() as model: # context management
           ...:     p = pm.Beta("p", alpha=alpha, beta=beta)
           ...:     y = pm.Binomial("y", n=n, p=p, observed=h)
           ...:     idata = pm.sample()

        In [3]: az.summary(idata, kind="stats")

        Out[3]:
            mean     sd  hdi_3%  hdi_97%
        p  0.609  0.047   0.528    0.699
    """
    if "start" in kwargs:
        if initvals is not None:
            raise ValueError("Passing both `start` and `initvals` is not supported.")
        warnings.warn(
            "The `start` kwarg was renamed to `initvals` and can now do more. Please check the docstring.",
            FutureWarning,
            stacklevel=2,
        )
        initvals = kwargs.pop("start")
    if nuts_sampler_kwargs is None:
        nuts_sampler_kwargs = {}
    if "target_accept" in kwargs:
        if "nuts" in kwargs and "target_accept" in kwargs["nuts"]:
            raise ValueError(
                "`target_accept` was defined twice. Please specify it either as a direct keyword argument or in the `nuts` kwarg."
            )
        if "nuts" in kwargs:
            kwargs["nuts"]["target_accept"] = kwargs.pop("target_accept")
        else:
            kwargs["nuts"] = {"target_accept": kwargs.pop("target_accept")}
    if isinstance(trace, list):
        raise ValueError("Please use `var_names` keyword argument for partial traces.")

    model = modelcontext(model)
    if not model.free_RVs:
        raise SamplingError(
            "Cannot sample from the model, since the model does not contain any free variables."
        )

    if cores is None:
        cores = min(4, _cpu_count())

    if chains is None:
        chains = max(2, cores)

    if blas_cores == "auto":
        blas_cores = cores

    cores = min(cores, chains)

    num_blas_cores_per_chain: int | None
    joined_blas_limiter: Callable[[], Any]

    if blas_cores is None:
        joined_blas_limiter = contextlib.nullcontext
        num_blas_cores_per_chain = None
    elif isinstance(blas_cores, int):

        def joined_blas_limiter():
            return threadpool_limits(limits=blas_cores)

        num_blas_cores_per_chain = blas_cores // cores
    else:
        raise ValueError(
            f"Invalid argument `blas_cores`, must be int, 'auto' or None: {blas_cores}"
        )

    if random_seed == -1:
        random_seed = None
    random_seed_list = _get_seeds_per_chain(random_seed, chains)

    if not discard_tuned_samples and not return_inferencedata:
        warnings.warn(
            "Tuning samples will be included in the returned `MultiTrace` object, which can lead to"
            " complications in your downstream analysis. Please consider to switch to `InferenceData`:\n"
            "`pm.sample(..., return_inferencedata=True)`",
            UserWarning,
            stacklevel=2,
        )

    # small trace warning
    if draws == 0:
        msg = "Tuning was enabled throughout the whole trace."
        _log.warning(msg)
    elif draws < 100:
        msg = f"Only {draws} samples per chain. Reliable r-hat and ESS diagnostics require longer chains for accurate estimate."
        _log.warning(msg)

    auto_nuts_init = True
    if step is not None:
        if isinstance(step, CompoundStep):
            for method in step.methods:
                if isinstance(method, NUTS):
                    auto_nuts_init = False
        elif isinstance(step, NUTS):
            auto_nuts_init = False

    initial_points = None
    step = assign_step_methods(model, step, methods=pm.STEP_METHODS, step_kwargs=kwargs)

    if nuts_sampler != "pymc":
        if not isinstance(step, NUTS):
            raise ValueError(
                "Model can not be sampled with NUTS alone. Your model is probably not continuous."
            )

        with joined_blas_limiter():
            return _sample_external_nuts(
                sampler=nuts_sampler,
                draws=draws,
                tune=tune,
                chains=chains,
                target_accept=kwargs.pop("nuts", {}).get("target_accept", 0.8),
                random_seed=random_seed,
                initvals=initvals,
                model=model,
                var_names=var_names,
                progressbar=progressbar,
                idata_kwargs=idata_kwargs,
                compute_convergence_checks=compute_convergence_checks,
                nuts_sampler_kwargs=nuts_sampler_kwargs,
                **kwargs,
            )

    if isinstance(step, list):
        step = CompoundStep(step)
    elif isinstance(step, NUTS) and auto_nuts_init:
        if "nuts" in kwargs:
            nuts_kwargs = kwargs.pop("nuts")
            [kwargs.setdefault(k, v) for k, v in nuts_kwargs.items()]
        _log.info("Auto-assigning NUTS sampler...")
        with joined_blas_limiter():
            initial_points, step = init_nuts(
                init=init,
                chains=chains,
                n_init=n_init,
                model=model,
                random_seed=random_seed_list,
                progressbar=progressbar,
                jitter_max_retries=jitter_max_retries,
                tune=tune,
                initvals=initvals,
                **kwargs,
            )

    if initial_points is None:
        # Time to draw/evaluate numeric start points for each chain.
        ipfns = make_initial_point_fns_per_chain(
            model=model,
            overrides=initvals,
            jitter_rvs=set(),
            chains=chains,
        )
        initial_points = [ipfn(seed) for ipfn, seed in zip(ipfns, random_seed_list)]

    # One final check that shapes and logps at the starting points are okay.
    ip: dict[str, np.ndarray]
    for ip in initial_points:
        model.check_start_vals(ip)
        _check_start_shape(model, ip)

    if var_names is not None:
        trace_vars = [v for v in model.unobserved_RVs if v.name in var_names]
        trace_vars = model.replace_rvs_by_values(trace_vars)
        assert len(trace_vars) == len(var_names), "Not all var_names were found in the model"
    else:
        trace_vars = None

    # Create trace backends for each chain
    run, traces = init_traces(
        backend=trace,
        chains=chains,
        expected_length=draws + tune,
        step=step,
        trace_vars=trace_vars,
        initial_point=ip,
        model=model,
    )

    sample_args = {
        # draws is now the total number of draws, including tuning
        "draws": draws + tune,
        "step": step,
        "start": initial_points,
        "traces": traces,
        "chains": chains,
        "tune": tune,
        "var_names": var_names,
        "progressbar": progressbar,
        "progressbar_theme": progressbar_theme,
        "model": model,
        "cores": cores,
        "callback": callback,
        "discard_tuned_samples": discard_tuned_samples,
    }
    parallel_args = {
        "mp_ctx": mp_ctx,
        "blas_cores": num_blas_cores_per_chain,
    }

    sample_args.update(kwargs)

    has_population_samplers = np.any(
        [
            isinstance(m, PopulationArrayStepShared)
            for m in (step.methods if isinstance(step, CompoundStep) else [step])
        ]
    )

    parallel = cores > 1 and chains > 1 and not has_population_samplers
    # At some point it was decided that PyMC should not set a global seed by default,
    # unless the user specified a seed. This is a symptom of the fact that PyMC samplers
    # are built around global seeding. This branch makes sure we maintain this unspoken
    # rule. See https://github.com/pymc-devs/pymc/pull/1395.
    if parallel:
        # For parallel sampling we can pass the list of random seeds directly, as
        # global seeding will only be called inside each process
        sample_args["random_seed"] = random_seed_list
    else:
        # We pass None if the original random seed was None. The single core sampler
        # methods will only set a global seed when it is not None.
        sample_args["random_seed"] = random_seed if random_seed is None else random_seed_list

    t_start = time.time()
    if parallel:
        _log.info(f"Multiprocess sampling ({chains} chains in {cores} jobs)")
        _print_step_hierarchy(step)
        try:
            _mp_sample(**sample_args, **parallel_args)
        except pickle.PickleError:
            _log.warning("Could not pickle model, sampling singlethreaded.")
            _log.debug("Pickling error:", exc_info=True)
            parallel = False
        except AttributeError as e:
            if not str(e).startswith("AttributeError: Can't pickle"):
                raise
            _log.warning("Could not pickle model, sampling singlethreaded.")
            _log.debug("Pickling error:", exc_info=True)
            parallel = False
    if not parallel:
        if has_population_samplers:
            _log.info(f"Population sampling ({chains} chains)")
            _print_step_hierarchy(step)
            with joined_blas_limiter():
                _sample_population(
                    initial_points=initial_points, parallelize=cores > 1, **sample_args
                )
        else:
            _log.info(f"Sequential sampling ({chains} chains in 1 job)")
            _print_step_hierarchy(step)
            with joined_blas_limiter():
                _sample_many(**sample_args)

    t_sampling = time.time() - t_start

    # Packaging, validating and returning the result was extracted
    # into a function to make it easier to test and refactor.
    return _sample_return(
        run=run,
        traces=traces,
        tune=tune,
        t_sampling=t_sampling,
        discard_tuned_samples=discard_tuned_samples,
        compute_convergence_checks=compute_convergence_checks,
        return_inferencedata=return_inferencedata,
        keep_warning_stat=keep_warning_stat,
        idata_kwargs=idata_kwargs or {},
        model=model,
    )


def _sample_return(
    *,
    run: RunType | None,
    traces: Sequence[IBaseTrace],
    tune: int,
    t_sampling: float,
    discard_tuned_samples: bool,
    compute_convergence_checks: bool,
    return_inferencedata: bool,
    keep_warning_stat: bool,
    idata_kwargs: dict[str, Any],
    model: Model,
) -> InferenceData | MultiTrace:
    """Final step of `pm.sampler` that picks/slices chains,
    runs diagnostics and converts to the desired return type."""
    # Pick and slice chains to keep the maximum number of samples
    if discard_tuned_samples:
        traces, length = _choose_chains(traces, tune)
    else:
        traces, length = _choose_chains(traces, 0)
    mtrace = MultiTrace(traces)[:length]

    # count the number of tune/draw iterations that happened
    # ideally via the "tune" statistic, but not all samplers record it!
    if "tune" in mtrace.stat_names:
        # Get the tune stat directly from chain 0, sampler 0
        stat = mtrace._straces[0].get_sampler_stats("tune", sampler_idx=0)
        stat = tuple(stat)
        n_tune = stat.count(True)
        n_draws = stat.count(False)
    else:
        # these may be wrong when KeyboardInterrupt happened, but they're better than nothing
        n_tune = min(tune, len(mtrace))
        n_draws = max(0, len(mtrace) - n_tune)

    if discard_tuned_samples:
        mtrace = mtrace[n_tune:]

    # save metadata in SamplerReport
    mtrace.report._n_tune = n_tune
    mtrace.report._n_draws = n_draws
    mtrace.report._t_sampling = t_sampling

    n_chains = len(mtrace.chains)
    _log.info(
        f'Sampling {n_chains} chain{"s" if n_chains > 1 else ""} for {n_tune:_d} tune and {n_draws:_d} draw iterations '
        f"({n_tune*n_chains:_d} + {n_draws*n_chains:_d} draws total) "
        f"took {t_sampling:.0f} seconds."
    )

    idata = None
    if compute_convergence_checks or return_inferencedata:
        ikwargs: dict[str, Any] = dict(model=model, save_warmup=not discard_tuned_samples)
        ikwargs.update(idata_kwargs)
        idata = pm.to_inference_data(mtrace, **ikwargs)

        if compute_convergence_checks:
            warns = run_convergence_checks(idata, model)
            mtrace.report._add_warnings(warns)
            log_warnings(warns)

        if return_inferencedata:
            # By default we drop the "warning" stat which contains `SamplerWarning`
            # objects that can not be stored with `.to_netcdf()`.
            if not keep_warning_stat:
                return drop_warning_stat(idata)
            return idata
    return mtrace


def _check_start_shape(model, start: PointType):
    """Checks that the prior evaluations and initial points have identical shapes.

    Parameters
    ----------
    model : pm.Model
        The current model on context.
    start : dict
        The complete dictionary mapping (transformed) variable names to numeric initial values.
    """
    e = ""
    try:
        actual_shapes = model.eval_rv_shapes()
    except NotImplementedError as ex:
        warnings.warn(f"Unable to validate shapes: {ex.args[0]}", UserWarning)
        return
    for name, sval in start.items():
        ashape = actual_shapes.get(name)
        sshape = np.shape(sval)
        if ashape != tuple(sshape):
            e += f"\nExpected shape {ashape} for var '{name}', got: {sshape}"
    if e != "":
        raise ValueError(f"Bad shape in start point:{e}")


def _sample_many(
    *,
    draws: int,
    chains: int,
    traces: Sequence[IBaseTrace],
    start: Sequence[PointType],
    random_seed: Sequence[RandomSeed] | None,
    step: Step,
    callback: SamplingIteratorCallback | None = None,
    **kwargs,
):
    """Samples all chains sequentially.

    Parameters
    ----------
    draws: int
        The number of samples to draw
    chains: int
        Total number of chains to sample.
    start: list
        Starting points for each chain
    random_seed: list of random seeds, optional
        A list of seeds, one for each chain
    step: function
        Step function
    """
    for i in range(chains):
        _sample(
            draws=draws,
            chain=i,
            start=start[i],
            step=step,
            trace=traces[i],
            random_seed=None if random_seed is None else random_seed[i],
            callback=callback,
            **kwargs,
        )
    return


def _sample(
    *,
    chain: int,
    progressbar: bool,
    random_seed: RandomSeed,
    start: PointType,
    draws: int,
    step: Step,
    trace: IBaseTrace,
    tune: int,
    model: Model | None = None,
    progressbar_theme: Theme | None = default_progress_theme,
    callback=None,
    **kwargs,
) -> None:
    """Main iteration for singleprocess sampling.

    Multiple step methods are supported via compound step methods.

    Parameters
    ----------
    chain : int
        Number of the chain that the samples will belong to.
    progressbar : bool
        Whether or not to display a progress bar in the command line. The bar shows the percentage
        of completion, the sampling speed in samples per second (SPS), and the estimated remaining
        time until completion ("expected time of arrival"; ETA).
    random_seed : single random seed
    start : dict
        Starting point in parameter space (or partial point)
    draws : int
        The number of samples to draw
    step : function
        Step function
    trace
        A chain backend to record draws and stats.
    tune : int
        Number of iterations to tune.
    model : Model (optional if in ``with`` context)
    progressbar_theme : Theme
        Optional custom theme for the progress bar.
    """
    skip_first = kwargs.get("skip_first", 0)

    sampling_gen = _iter_sample(
        draws=draws,
        step=step,
        start=start,
        trace=trace,
        chain=chain,
        tune=tune,
        model=model,
        random_seed=random_seed,
        callback=callback,
    )
    _pbar_data = {"chain": chain, "divergences": 0}
    _desc = "Sampling chain {chain:d}, {divergences:,d} divergences"

    progress = CustomProgress(
        "[progress.description]{task.description}",
        BarColumn(),
        "[progress.percentage]{task.percentage:>3.0f}%",
        TimeRemainingColumn(),
        TextColumn("/"),
        TimeElapsedColumn(),
        console=Console(theme=progressbar_theme),
        disable=not progressbar,
    )

    with progress:
        try:
            task = progress.add_task(_desc.format(**_pbar_data), completed=0, total=draws)
            for it, diverging in enumerate(sampling_gen):
                if it >= skip_first and diverging:
                    _pbar_data["divergences"] += 1
                progress.update(task, description=_desc.format(**_pbar_data), completed=it)
            progress.update(
                task, description=_desc.format(**_pbar_data), completed=draws, refresh=True
            )
        except KeyboardInterrupt:
            pass


def _iter_sample(
    *,
    draws: int,
    step: Step,
    start: PointType,
    trace: IBaseTrace,
    chain: int = 0,
    tune: int = 0,
    model: Model | None = None,
    random_seed: RandomSeed = None,
    callback: SamplingIteratorCallback | None = None,
) -> Iterator[bool]:
    """Generator for sampling one chain. (Used in singleprocess sampling.)

    Parameters
    ----------
    draws : int
        The number of samples to draw
    step : function
        Step function
    start : dict
        Starting point in parameter space (or partial point).
        Must contain numeric (transformed) initial values for all (transformed) free variables.
    trace
        A chain backend to record draws and stats.
    chain : int, optional
        Chain number used to store sample in backend.
    tune : int, optional
        Number of iterations to tune (defaults to 0).
    model : Model (optional if in ``with`` context)
    random_seed : single random seed, optional

    Yields
    ------
    diverging : bool
        Indicates if the draw is divergent. Only available with some samplers.
    """
    model = modelcontext(model)
    draws = int(draws)

    if draws < 1:
        raise ValueError("Argument `draws` must be greater than 0.")

    if random_seed is not None:
        np.random.seed(random_seed)

    point = start

    try:
        step.tune = bool(tune)
        if hasattr(step, "reset_tuning"):
            step.reset_tuning()
        for i in range(draws):
            diverging = False

            if i == 0 and hasattr(step, "iter_count"):
                step.iter_count = 0
            if i == tune:
                step.stop_tuning()
            point, stats = step.step(point)
            trace.record(point, stats)
            log_warning_stats(stats)
            diverging = i > tune and len(stats) > 0 and (stats[0].get("diverging") is True)
            if callback is not None:
                callback(
                    trace=trace,
                    draw=Draw(chain, i == draws, i, i < tune, stats, point),
                )

            yield diverging
    except KeyboardInterrupt:
        trace.close()
        raise
    except BaseException:
        trace.close()
        raise
    else:
        trace.close()


def _mp_sample(
    *,
    draws: int,
    tune: int,
    step,
    chains: int,
    cores: int,
    random_seed: Sequence[RandomSeed],
    start: Sequence[PointType],
    progressbar: bool = True,
    progressbar_theme: Theme | None = default_progress_theme,
    traces: Sequence[IBaseTrace],
    model: Model | None = None,
    callback: SamplingIteratorCallback | None = None,
    blas_cores: int | None = None,
    mp_ctx=None,
    **kwargs,
) -> None:
    """Main iteration for multiprocess sampling.

    Parameters
    ----------
    draws : int
        The number of samples to draw
    tune : int
        Number of iterations to tune.
    step : function
        Step function
    chains : int
        The number of chains to sample.
    cores : int
        The number of chains to run in parallel.
    random_seed : list of random seeds
        Random seeds for each chain.
    start : list
        Starting points for each chain.
        Dicts must contain numeric (transformed) initial values for all (transformed) free variables.
    progressbar : bool
        Whether or not to display a progress bar in the command line.
    progressbar_theme : Theme
        Optional custom theme for the progress bar.
    traces
        Recording backends for each chain.
    model : Model (optional if in ``with`` context)
    callback
        A function which gets called for every sample from the trace of a chain. The function is
        called with the trace and the current draw and will contain all samples for a single trace.
        the ``draw.chain`` argument can be used to determine which of the active chains the sample
        is drawn from.
        Sampling can be interrupted by throwing a ``KeyboardInterrupt`` in the callback.
    """
    import pymc.sampling.parallel as ps

    # We did draws += tune in pm.sample
    draws -= tune

    sampler = ps.ParallelSampler(
        draws=draws,
        tune=tune,
        chains=chains,
        cores=cores,
        seeds=random_seed,
        start_points=start,
        step_method=step,
        progressbar=progressbar,
        progressbar_theme=progressbar_theme,
        blas_cores=blas_cores,
        mp_ctx=mp_ctx,
    )
    try:
        try:
            with sampler:
                for draw in sampler:
                    strace = traces[draw.chain]
                    strace.record(draw.point, draw.stats)
                    log_warning_stats(draw.stats)
                    if draw.is_last:
                        strace.close()

                    if callback is not None:
                        callback(trace=strace, draw=draw)

        except ps.ParallelSamplingError as error:
            strace = traces[error._chain]
            for strace in traces:
                strace.close()
            raise
    except KeyboardInterrupt:
        pass
    finally:
        for strace in traces:
            strace.close()


def _init_jitter(
    model: Model,
    initvals: StartDict | Sequence[StartDict | None] | None,
    seeds: Sequence[int] | np.ndarray,
    jitter: bool,
    jitter_max_retries: int,
) -> list[PointType]:
    """Apply a uniform jitter in [-1, 1] to the test value as starting point in each chain.

    ``model.check_start_vals`` is used to test whether the jittered starting
    values produce a finite log probability. Invalid values are resampled
    unless `jitter_max_retries` is achieved, in which case the last sampled
    values are returned.

    Parameters
    ----------
    jitter: bool
        Whether to apply jitter or not.
    jitter_max_retries : int
        Maximum number of repeated attempts at initializing values (per chain).

    Returns
    -------
    start : ``pymc.model.Point``
        Starting point for sampler
    """

    ipfns = make_initial_point_fns_per_chain(
        model=model,
        overrides=initvals,
        jitter_rvs=set(model.free_RVs) if jitter else set(),
        chains=len(seeds),
    )

    if not jitter:
        return [ipfn(seed) for ipfn, seed in zip(ipfns, seeds)]

    initial_points = []
    for ipfn, seed in zip(ipfns, seeds):
        rng = np.random.RandomState(seed)
        for i in range(jitter_max_retries + 1):
            point = ipfn(seed)
            if i < jitter_max_retries:
                try:
                    model.check_start_vals(point)
                except SamplingError:
                    # Retry with a new seed
                    seed = rng.randint(2**30, dtype=np.int64)
                else:
                    break
        initial_points.append(point)
    return initial_points


def init_nuts(
    *,
    init: str = "auto",
    chains: int = 1,
    n_init: int = 500_000,
    model: Model | None = None,
    random_seed: RandomSeed = None,
    progressbar=True,
    jitter_max_retries: int = 10,
    tune: int | None = None,
    initvals: StartDict | Sequence[StartDict | None] | None = None,
    **kwargs,
) -> tuple[Sequence[PointType], NUTS]:
    """Set up the mass matrix initialization for NUTS.

    NUTS convergence and sampling speed is extremely dependent on the
    choice of mass/scaling matrix. This function implements different
    methods for choosing or adapting the mass matrix.

    Parameters
    ----------
    init : str
        Initialization method to use.

        * auto: Choose a default initialization method automatically.
          Currently, this is ``jitter+adapt_diag``, but this can change in the future. If you
          depend on the exact behaviour, choose an initialization method explicitly.
        * adapt_diag: Start with a identity mass matrix and then adapt a diagonal based on the
          variance of the tuning samples. All chains use the test value (usually the prior mean)
          as starting point.
        * jitter+adapt_diag: Same as ``adapt_diag``, but use test value plus a uniform jitter in
          [-1, 1] as starting point in each chain.
        * jitter+adapt_diag_grad:
          An experimental initialization method that uses information from gradients and samples
          during tuning.
        * advi+adapt_diag: Run ADVI and then adapt the resulting diagonal mass matrix based on the
          sample variance of the tuning samples.
        * advi: Run ADVI to estimate posterior mean and diagonal mass matrix.
        * advi_map: Initialize ADVI with MAP and use MAP as starting point.
        * map: Use the MAP as starting point. This is discouraged.
        * adapt_full: Adapt a dense mass matrix using the sample covariances. All chains use the
          test value (usually the prior mean) as starting point.
        * jitter+adapt_full: Same as ``adapt_full``, but use test value plus a uniform jitter in
          [-1, 1] as starting point in each chain.

    chains : int
        Number of jobs to start.
    initvals : optional, dict or list of dicts
        Dict or list of dicts with initial values to use instead of the defaults from `Model.initial_values`.
        The keys should be names of transformed random variables.
    n_init : int
        Number of iterations of initializer. Only works for 'ADVI' init methods.
    model : Model (optional if in ``with`` context)
    random_seed : int, array-like of int, RandomState or Generator, optional
        Seed for the random number generator.
    progressbar : bool
        Whether or not to display a progressbar for advi sampling.
    jitter_max_retries : int
        Maximum number of repeated attempts (per chain) at creating an initial matrix with uniform jitter
        that yields a finite probability. This applies to ``jitter+adapt_diag`` and ``jitter+adapt_full``
        init methods.
    **kwargs : keyword arguments
        Extra keyword arguments are forwarded to pymc.NUTS.

    Returns
    -------
    initial_points : list
        Starting points for each chain.
    nuts_sampler : ``pymc.step_methods.NUTS``
        Instantiated and initialized NUTS sampler object
    """
    model = modelcontext(model)

    vars = kwargs.get("vars", model.value_vars)
    if set(vars) != set(model.value_vars):
        raise ValueError("Must use init_nuts on all variables of a model.")
    if not all_continuous(vars):
        raise ValueError("init_nuts can only be used for models with continuous variables.")

    if not isinstance(init, str):
        raise TypeError("init must be a string.")

    init = init.lower()

    if init == "auto":
        init = "jitter+adapt_diag"

    random_seed_list = _get_seeds_per_chain(random_seed, chains)

    _log.info(f"Initializing NUTS using {init}...")

    cb = [
        pm.callbacks.CheckParametersConvergence(tolerance=1e-2, diff="absolute"),
        pm.callbacks.CheckParametersConvergence(tolerance=1e-2, diff="relative"),
    ]

    initial_points = _init_jitter(
        model,
        initvals,
        seeds=random_seed_list,
        jitter="jitter" in init,
        jitter_max_retries=jitter_max_retries,
    )

    apoints = [DictToArrayBijection.map(point) for point in initial_points]
    apoints_data = [apoint.data for apoint in apoints]
    potential: quadpotential.QuadPotential

    if init == "adapt_diag":
        mean = np.mean(apoints_data, axis=0)
        var = np.ones_like(mean)
        n = len(var)
        potential = quadpotential.QuadPotentialDiagAdapt(n, mean, var, 10)
    elif init == "jitter+adapt_diag":
        mean = np.mean(apoints_data, axis=0)
        var = np.ones_like(mean)
        n = len(var)
        potential = quadpotential.QuadPotentialDiagAdapt(n, mean, var, 10)
    elif init == "jitter+adapt_diag_grad":
        mean = np.mean(apoints_data, axis=0)
        var = np.ones_like(mean)
        n = len(var)

        if tune is not None and tune > 250:
            stop_adaptation = tune - 50
        else:
            stop_adaptation = None

        potential = quadpotential.QuadPotentialDiagAdaptExp(
            n,
            mean,
            alpha=0.02,
            use_grads=True,
            stop_adaptation=stop_adaptation,
        )
    elif init == "advi+adapt_diag":
        approx = pm.fit(
            random_seed=random_seed_list[0],
            n=n_init,
            method="advi",
            model=model,
            callbacks=cb,
            progressbar=progressbar,
            obj_optimizer=pm.adagrad_window,
        )
        approx_sample = approx.sample(
            draws=chains, random_seed=random_seed_list[0], return_inferencedata=False
        )
        initial_points = [approx_sample[i] for i in range(chains)]
        std_apoint = approx.std.eval()
        cov = std_apoint**2
        mean = approx.mean.get_value()
        weight = 50
        n = len(cov)
        potential = quadpotential.QuadPotentialDiagAdapt(n, mean, cov, weight)
    elif init == "advi":
        approx = pm.fit(
            random_seed=random_seed_list[0],
            n=n_init,
            method="advi",
            model=model,
            callbacks=cb,
            progressbar=progressbar,
            obj_optimizer=pm.adagrad_window,
        )
        approx_sample = approx.sample(
            draws=chains, random_seed=random_seed_list[0], return_inferencedata=False
        )
        initial_points = [approx_sample[i] for i in range(chains)]
        cov = approx.std.eval() ** 2
        potential = quadpotential.QuadPotentialDiag(cov)
    elif init == "advi_map":
        start = pm.find_MAP(include_transformed=True, seed=random_seed_list[0])
        approx = pm.MeanField(model=model, start=start)
        pm.fit(
            random_seed=random_seed_list[0],
            n=n_init,
            method=pm.KLqp(approx),
            callbacks=cb,
            progressbar=progressbar,
            obj_optimizer=pm.adagrad_window,
        )
        approx_sample = approx.sample(
            draws=chains, random_seed=random_seed_list[0], return_inferencedata=False
        )
        initial_points = [approx_sample[i] for i in range(chains)]
        cov = approx.std.eval() ** 2
        potential = quadpotential.QuadPotentialDiag(cov)
    elif init == "map":
        start = pm.find_MAP(include_transformed=True, seed=random_seed_list[0])
        cov = -pm.find_hessian(point=start, negate_output=False)
        initial_points = [start] * chains
        potential = quadpotential.QuadPotentialFull(cov)
    elif init == "adapt_full":
        mean = np.mean(apoints_data * chains, axis=0)
        initial_point = initial_points[0]
        initial_point_model_size = sum(initial_point[n.name].size for n in model.value_vars)
        cov = np.eye(initial_point_model_size)
        potential = quadpotential.QuadPotentialFullAdapt(initial_point_model_size, mean, cov, 10)
    elif init == "jitter+adapt_full":
        mean = np.mean(apoints_data, axis=0)
        initial_point = initial_points[0]
        initial_point_model_size = sum(initial_point[n.name].size for n in model.value_vars)
        cov = np.eye(initial_point_model_size)
        potential = quadpotential.QuadPotentialFullAdapt(initial_point_model_size, mean, cov, 10)
    else:
        raise ValueError(f"Unknown initializer: {init}.")

    step = pm.NUTS(potential=potential, model=model, **kwargs)

    # Filter deterministics from initial_points
    value_var_names = [var.name for var in model.value_vars]
    initial_points = [
        {k: v for k, v in initial_point.items() if k in value_var_names}
        for initial_point in initial_points
    ]

    return initial_points, step



# coding: utf-8
"""
    MockMain.py

    Copyright (c) 2020, SAXS Team, KEK-PF
"""
import logging
from molass_legacy.KekLib.OurTkinter import Tk, Dialog
from molass_legacy.KekLib.BasicUtils import Struct

class MockMain(Dialog):
    def __init__(self, parent):
        self.logger = logging.getLogger(__name__)
        self.testing = False
        self.analyzer = Struct(app_logger=self.logger)
        Dialog.__init__(self, parent, title="Mock Main", visible=False)

    def show(self):
        self._show()



# coding: utf-8
"""

    ScatteringBaseInvestigator.py

    Copyright (c) 2017-2019, SAXS Team, KEK-PF

"""
from molass_legacy.KekLib.OurTkinter             import Tk, Dialog
from molass_legacy.KekLib.TkSupplements          import tk_set_icon_portable
from molass_legacy.KekLib.TkUtils                import is_low_resolution
from CanvasFrame            import CanvasFrame

class InvestigatorDialog( Dialog ):
    def __init__( self, parent, corrector ):
        self.parent     = parent
        self.corrector  = corrector
        self.max_index  = len(self.corrector.qvector)-1

    def show( self ):
        title = "ScatteringBaseInvestigator"
        Dialog.__init__(self, self.parent, title )

    def body( self, body_frame ):   # overrides parent class method
        tk_set_icon_portable( self )

        base_frame = Tk.Frame( body_frame );
        base_frame.pack( expand=1, fill=Tk.BOTH, padx=20, pady=10 )

        figsize = ( 16, 8 ) if is_low_resolution() else ( 16, 8 )
        self.canvas_frame = canvas_frame = CanvasFrame( base_frame, figsize=figsize )
        canvas_frame.pack()

        fig = canvas_frame.fig
        ax1 = fig.add_subplot( 121 )
        ax2 = fig.add_subplot( 122 )
        self.axes = [ ax1, ax2 ]

        self.index_var  = Tk.IntVar()
        self.index_var.set( 0 )
        self.inc_var    = Tk.IntVar()
        self.inc_var.set( 10 )
        spinbox_frame = Tk.Frame( base_frame )
        spinbox_frame.pack()

        label = Tk.Label( spinbox_frame, text="Investigating " )
        label.grid( row=0, column=0 )
        self.spinbox = Tk.Spinbox( spinbox_frame, textvariable=self.index_var,
                                from_=0, to=self.max_index, increment=self.inc_var.get(), 
                                justify=Tk.CENTER, width=6 )
        self.spinbox.grid( row=0, column=1 )
        label = Tk.Label( spinbox_frame, text="-th Q-Plane with spinbox increment" )
        label.grid( row=0, column=2 )
        inc_entry = Tk.Entry( spinbox_frame, textvariable=self.inc_var, justify=Tk.CENTER, width=3 )
        inc_entry.grid( row=0, column=3 )
        label = Tk.Label( spinbox_frame, text=" and plot options " )
        label.grid( row=0, column=4 )
        self.supress_clear = Tk.IntVar()
        self.supress_clear.set( 0 )
        cb = Tk.Checkbutton( spinbox_frame, text="without clearing", variable=self.supress_clear )
        cb.grid( row=0, column=5 )

        self.same_scaling = Tk.IntVar()
        self.same_scaling.set( 1 )
        cb = Tk.Checkbutton( spinbox_frame, text="same scaling", variable=self.same_scaling )
        cb.grid( row=0, column=6 )

        self.zero_base = Tk.IntVar()
        self.zero_base.set( 0 )
        cb = Tk.Checkbutton( spinbox_frame, text="zero base", variable=self.zero_base )
        cb.grid( row=0, column=7 )

        self.index_var.trace( 'w', self.index_var_tracer )
        self.index_var_tracer()
        self.inc_var.trace( 'w', self.inc_var_tracer )
        self.same_scaling.trace( 'w', self.redraw_tracer )
        self.zero_base.trace( 'w', self.redraw_tracer )

    def buttonbox(self):
        '''add standard button box.

        override if you do not want the standard buttons
        '''

        box = Tk.Frame(self)

        w = Tk.Button(box, text="Back", width=10, command=self.back_, default=Tk.ACTIVE)
        w.pack(side=Tk.LEFT, padx=5, pady=5)
        self.back_button = w
        w = Tk.Button(box, text="Next", width=10, command=self.next_, default=Tk.ACTIVE)
        w.pack(side=Tk.LEFT, padx=5, pady=5)
        self.next_button = w
        w = Tk.Button(box, text="Cancel", width=10, command=self.cancel)
        w.pack(side=Tk.LEFT, padx=5, pady=5)
        self.cancel_button = w

        self.bind("<Escape>", self.cancel)

        box.pack()

    def index_var_tracer( self, *args ):
        try:
            i   = self.index_var.get()
        except:
            # possibly when ""
            return

        if i >= 0 and i <= self.max_index:
            self.draw_a_plane( i )
        else:
            # better be silent
            pass

    def inc_var_tracer( self, *args ):
        try:
            inc = self.inc_var.get()
        except:
            return

        if inc > 0 and inc < self.max_index//2:
            self.spinbox.config( increment=inc )

    def redraw_tracer( self, *args ):
        try:
            i   = self.index_var.get()
        except:
            # possibly when ""
            return

        self.draw_a_plane( max( 0, min( self.max_index, i ) ) )

    def next_( self ):
        try:
            i   = self.index_var.get()
            inc = self.inc_var.get()
        except:
            return
        self.index_var.set( max( 0, min( self.max_index, i + inc ) ) )

    def back_( self ):
        try:
            i   = self.index_var.get()
            inc = self.inc_var.get()
        except:
            return
        self.index_var.set( max( 0, min( self.max_index, i - inc ) ) )

    def draw_a_plane( self, i ):
        plot_closure = self.corrector.correct_a_single_q_plane( i, suppress_update=True, plot_always=True )
        if plot_closure is None:
            return

        self.axes[0].cla()
        if self.supress_clear.get() == 0:
            self.axes[1].cla()

        same_scaling    = self.same_scaling.get() == 1
        zero_base       = self.zero_base.get() == 1
        plot_closure( self.canvas_frame.fig, axes=self.axes, same_scaling=same_scaling, zero_base=zero_base )
        self.canvas_frame.show()



"""
    GuinierTools.CpdUtils.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import numpy as np
import ruptures as rpt

def compute_end_points(nc, x_, rgv):
    print("cpd_spike_impl")

    algo = rpt.Dynp(model="l2").fit(rgv)
    my_bkps = algo.predict(n_bkps=nc-1)
    print("my_bkps=", my_bkps)
    points = np.concatenate([[0], my_bkps[:-1], [len(x_)-1]])
    return x_[points].copy()


# coding: utf-8
"""
    ConcNormalizer.py

    Copyright (c) 2019, SAXS Team, KEK-PF
"""
import os
import glob
from SerialDataUtils import load_xray_files
from molass_legacy.KekLib.NumpyUtils import np_savetxt_with_comments

def find_conc_file(in_folder):
    files = glob.glob(in_folder + "/Conc*.txt")
    if len(files) > 0:
        _, ret_file = os.path.split(files[0])
    else:
        ret_file = None
    return ret_file

def normalize_impl(in_files, conc_file, out_folder, name_changer, dialog=None):
    conc_data = []
    with open(conc_file) as fh:
        for line in fh.readlines():
            conc_data.append(float(line))

    data_array, comments = load_xray_files(in_files)

    comments.insert(-2, "### comments from concetration normalization begin ###\n")
    comments.insert(-2, "# normalized with %s\n" % conc_file)
    comments.insert(-2, "### comments from concetration normalization end ###\n")

    if dialog is not None:
        dialog.mpb["maximum"] = len(in_files)
        dialog.mpb["value"] = 0

    for k in range(len(in_files)):
        data = data_array[k,:,:]
        data[:,1:] /= conc_data[k]
        _, file = os.path.split(in_files[k])
        file_ = name_changer(file)
        out_file = os.path.join(out_folder, file_)
        # print([k], out_file)
        np_savetxt_with_comments(out_file, data, comments)

        if dialog is not None:
            dialog.num_files_o.set(k+1)
            dialog.mpb["value"] = k+1
            dialog.update()



# coding: utf-8
"""

    ファイル名：   TextShowDialog.py

    処理内容：

        

"""
from __future__ import division, print_function, unicode_literals

from molass_legacy.KekLib.OurTkinter         import Tk, Dialog
from molass_legacy.KekLib.TkSupplements      import tk_set_icon_portable

class TextShowDialog( Dialog ):

    def __init__( self, parent, title, message='', width=80, height=50 ):
        self.grab = 'local'     # used in grab_set
        self.applied            = False
        self.message            = message
        self.width              = width
        self.height             = height

        Dialog.__init__(self, parent, title) # this calls body

    def body( self, body_frame ):   # overrides parent class method

        tk_set_icon_portable( self )

        iframe = Tk.Frame( body_frame );
        iframe.pack( expand=1, fill=Tk.BOTH, padx=10, pady=10 )

        message = Tk.Text( iframe, width=self.width, height=self.height )
        message.pack( expand=1 )
        message.insert( Tk.INSERT, self.message )
        # make it read-only
        message.config( state=Tk.DISABLED  )

        # global grab cannot be set befor windows is 'viewable'
        # and this happen in mainloop after this function returns
        # Thus, it is needed to delay grab setting of an interval
        # long enough to make sure that the window has been made
        # 'viewable'
        if self.grab == 'global':
            self.after(100, self.grab_set_global )
        else:
            pass # local grab is set by parent class constructor

    def buttonbox(self):
        '''
        overrides the standard buttons
        '''

        box = Tk.Frame(self)

        w = Tk.Button(box, text="OK", width=10, command=self.ok, default=Tk.ACTIVE)
        w.pack(side=Tk.LEFT, padx=5, pady=5)

        self.bind("<Return>", self.ok)

        box.pack( pady=10 )

    def apply( self ):  # overrides parent class method
        print( "ok" )



# -*- coding: utf-8 -*-
"""

    ファイル名：   OurNavigationToolbar.py

    処理内容：

       NavigationToolbar2TkAgg のカスタマイズ（Windows のみ）
       Ubuntu では元の NavigationToolbar2TkAgg で OK

    Copyright (c) 2016-2019, Masatsuyo Takahashi, KEK-PF

"""
from __future__ import division, print_function, unicode_literals
import os
import sys
import numpy as np
import pylab as pl
import tkinter as Tk
import matplotlib

if matplotlib.__version__ >= '2.2':
    from matplotlib.backends.backend_tkagg import NavigationToolbar2Tk
else:
    from matplotlib.backends.backend_tkagg import NavigationToolbar2TkAgg as NavigationToolbar2Tk

class NavigationToolbar( NavigationToolbar2Tk ):
    def __init__( self, canvas, window, show_mode=True ):
        self.show_mode = show_mode
        NavigationToolbar2Tk.__init__( self, canvas, window )

    """
        matplotlib.backend_bases.NavigationToolbar2.mouse_move を override
    """
    def mouse_move(self, event):
        self._set_cursor(event)

        if event.inaxes and event.inaxes.get_navigate():

            try:
                s = event.inaxes.format_coord(event.xdata, event.ydata)
            except (ValueError, OverflowError):
                pass
            else:

                # s += している部分を削除

                if self.show_mode and len(self.mode):
                    self.set_message('%s, %s' % (self.mode, s))
                else:
                    self.set_message(s)
        else:
            if self.show_mode:
                self.set_message(self.mode)
            else:
                self.set_message('')



# coding: utf-8
"""
    SharedArrays.py

    Creating multiple shared memories seems unstable as of
    Python 3.8.0 + numpy 1.17.4 + matplotlib 3.1.2

    Therefore, this module uses only one shared memory at a time,
    which contains multiple numpy arrays.

    Copyright (c) 2019, SAXS Team, KEK-PF
"""
import copy
import numpy as np

from multiprocessing.shared_memory import SharedMemory

NBYTES_FLOAT64 = 8

class SharedArrays:
    def __init__(self, arrays=None, tuples=None, name=None):
        if arrays is None:
            assert tuples is not None
            total_size = np.sum([np.prod(shape) * NBYTES_FLOAT64 for shape, dtype in tuples])
            existing_shm = SharedMemory(name=name, size=total_size)
            self.arrays = []
            self.tuples = copy.deepcopy(tuples)
            start = 0
            for shape, dtype in tuples:
                size = np.prod(shape) * NBYTES_FLOAT64
                stop = start + size
                sh_array = np.ndarray(shape, dtype=dtype, buffer=existing_shm.buf[start:stop])

                # deepcopy as a temporary fix to a suspected bug as of Python 3.8.0 + numpy 1.17.4 + matplotlib 3.1.2
                # therefore, these arrays should be used as read only
                self.arrays.append(copy.deepcopy(sh_array))

                start = stop
            self.shm = existing_shm
            self.name = name
        else:
            total_size = int(np.sum([a.nbytes for a in arrays]))
            shm = SharedMemory(create=True, size=total_size)
            name = shm.name
            self.arrays = []
            self.tuples = []
            start = 0
            for array in arrays:
                size = array.nbytes
                stop = start + size
                sh_array = np.ndarray(array.shape, dtype=array.dtype, buffer=shm.buf[start:stop])
                sh_array[:] = array[:]  # Copy the original data into shared memory
                self.arrays.append(sh_array)
                self.tuples.append((array.shape, array.dtype))
                start = stop
            self.shm = shm
            self.name = shm.name

    def get_name(self):
        return self.name

    def get_arrays(self):
        return self.arrays

    def get_tuples(self):
        return self.tuples



"""
    Theory.DjKinning1984.py

    Copyright (c) 2021, SAXS Team, KEK-PF
"""
import numpy as np
import molass_legacy.KekLib.DebugPlot as plt
from .SolidSphere import phi

def S0(q, R, K=1):
    return 1 - K*phi(q, 2*R)

def P0(q, R):
    return phi(q, R)**2

def S1(q, R, K=1):
    return 1/(1 + K*phi(q, 2*R))

def P1(q, R, K=1):
    return phi(q, R)**2 * S1(q, R, K=K)

def S(q, R, n):
    A = 2*q*R
    return 1/(1 + 24*n*(G(A, n)/A))

def G(A, n):
    n4 = (1 - n)**4
    alpha = (1 + 2*n)**2/n4
    beta = -6*n*(1 + n/2)**2/n4
    gamma = n*alpha/2

    return ( alpha*(np.sin(A) - A*np.cos(A))/A**2
            + beta*(2*A*np.sin(A) + (2 - A**2)*np.cos(A) - 2)/A**3
            + gamma*(-A**4*np.cos(A) + 4*((3*A**2 - 6)*np.cos(A) + (A**3 - 6*A)*np.sin(A) + 6))/A**5
            )

def P2(q, R, c=0):
    return phi(q, R)**2 * S(q, R, c)

def demo1():
    from .SolidSphere import get_boundary_params_simple

    qv = np.linspace(0.005, 0.3, 100)
    Rg = 35
    R = np.sqrt(5/3)*Rg

    p = P0(qv, R)
    s_ = np.ones(len(qv))
    s0 = S0(qv, R)
    s1 = S1(qv, R)

    b1, b2, k = get_boundary_params_simple(Rg)

    fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(16,7))
    fig.suptitle("Comparison of Structure Models for Solid Spheres", fontsize=20)

    ax1.set_title("Intensities: I(q)=P(q)*S(q)", fontsize=16)
    ax2.set_title("Structure Factors: S(q)", fontsize=16)

    ax1.set_yscale('log')

    label = 'S=1'
    ax1.plot(qv, p, label=label)
    ax2.plot(qv, s_, label=label)
    label = 'Debye (K=1)'
    ax1.plot(qv, p*s0, label=label)
    ax2.plot(qv, s0, label=label)
    label = 'Fournet (K=1)'
    ax1.plot(qv, p*s1, label=label)
    ax2.plot(qv, s1, label=label)

    for c in [0, 0.1, 0.2]:
        s2 = S(qv, R, c)
        label = 'Kinning (c=%.g)'%c
        ax1.plot(qv, p*s2, ':', label=label)
        ax2.plot(qv, s2, ':', label=label)

    for ax in [ax1, ax2]:
        ymin, ymax = ax.get_ylim()
        ax.set_ylim(ymin, ymax)
        for j, b in enumerate([b1, b2]):
            ax.plot([b, b], [ymin, ymax], ':', color='gray', label='b%d' % (j+1))

    ax1.legend()
    ax2.legend()
    fig.tight_layout()
    fig.subplots_adjust(top=0.85)
    plt.show()



"""
    Peaks/LrfRgComputer.py

    Copyright (c) 2024-2025, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.optimize import minimize
from scipy.interpolate import UnivariateSpline
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy.Models.ModelUtils import compute_cy_list, compute_area_props, get_paired_ranges_from_params_array

def compute_rgs_from_lrf_source(lrf_src, keep_num_components=False, want_num_components=None, select=None, debug=True):
    # task: unify this with Selective.MinimizerUtils.RgComputer
    # task: change get_paired_ranges_from_params_array into a class
    if debug:
        from importlib import reload
        import molass_legacy.GuinierTools.RgComputer
        reload(molass_legacy.GuinierTools.RgComputer)
    from molass_legacy.GuinierTools.RgComputer import compute_rgs

    print("compute_rgs_from_lrf_source")

    x = lrf_src.xr_x
    model = lrf_src.model
    cy_list = compute_cy_list(model, x, lrf_src.xr_peaks)
    props = compute_area_props(cy_list)
    if debug: 
        print("props=", props)

    xr_peaks = lrf_src.xr_peaks
    num_peaks = len(xr_peaks)
    M, E, qv, xr_curve = lrf_src.corrected_sd.get_xr_data_separate_ly()
    paired_ranges, indeces = get_paired_ranges_from_params_array(model, x, xr_peaks, return_indeces=True, want_num_components=want_num_components, select=select)
    C = np.array(cy_list)[indeces,:]    # note that, by indexing with "indeces", C does not include ignored components
    try:
        rgs, trs, peak_rgs, peak_trs, qualities = compute_rgs(M, E, qv, C, paired_ranges, x=x, return_trs=True, return_qualities=True, debug=debug)
    except AssertionError:
        from molass_legacy.KekLib.ExceptionTracebacker import log_exception
        log_exception(None, "compute_rgs: ")
        print("indeces=", indeces)
        print("C.shape=", C.shape)
        print("len(paired_ranges)=", len(paired_ranges))
        y = lrf_src.xr_y
        with plt.Dp():
            fig, ax = plt.subplots()
            ax.set_title("compute_rgs AssertionError")
            ax.plot(x, y)
            for k, cy in enumerate(cy_list):
                ax.plot(x, cy, ":", label="component-%d" % k)
            for k, pr in enumerate(paired_ranges):
                for f, t in pr.get_fromto_list():
                    print([k], f, t)
                    ax.axvspan(x[f], x[t], alpha=0.3, color="C%d" % k)
            ax.legend()
            fig.tight_layout() 
            plt.show()

    if keep_num_components:
        if len(paired_ranges) < num_peaks:
            import logging
            # compute minor rgs according to the exclusion curve
            # because they could be unreliable if computed directly by LRF
            full_paired_ranges = get_paired_ranges_from_params_array(model, x, xr_peaks, ignoreable_ratio=0)
            assert len(full_paired_ranges) == num_peaks
            spline = make_exclcurve_spline(rgs, trs)
            full_indeces = np.arange(num_peaks)
            minor_indeces = np.setdiff1d(full_indeces, indeces)
            new_peak_rgs = np.zeros(num_peaks)
            new_peak_trs = np.zeros(num_peaks)
            new_qualities = np.zeros(num_peaks)
            new_peak_rgs[indeces] = peak_rgs
            new_peak_trs[indeces] = peak_trs
            new_qualities[indeces] = qualities
            for i in minor_indeces:
                paired_range = full_paired_ranges[i]
                ft_list = paired_range.get_fromto_list()
                if len(ft_list) == 1:
                    tr = x[0] + np.average(ft_list[0])
                else:
                    tr = x[0] + ft_list[0][1]
                rg = spline(tr)
                new_peak_rgs[i] = rg
                new_peak_trs[i] = tr

            logger = logging.getLogger(__name__)
            logger.warning("Minor Rgs at %s in %s have been computed according to the exclusion curve.", str(minor_indeces), str(full_indeces))

            peak_rgs = new_peak_rgs
            peak_trs = new_peak_trs
            qualities = new_qualities
            indeces = full_indeces
        
    if debug:
        print("rgs=", rgs, "trs=", trs)

    ret_props = props[indeces]
    return rgs, trs, props, peak_rgs, peak_trs, ret_props/np.sum(ret_props), indeces, qualities

def make_exclcurve_spline(rgs, trs, debug=False):
    def objective(p):
        N, T, x0, poresize = p
        rhov = rgs/poresize
        rhov[rhov > 1] = 1
        trs_ = x0 + N*T*np.power(1 - rhov, 3)
        return np.sum((trs - trs_)**2)
    
    init_params = [2000, 1, 0, 200]
    res = minimize(objective, init_params, method='Nelder-Mead')
    N, T, x0, poresize = res.x
    y = np.linspace(poresize, 10, 100)
    rhov = y/poresize   # note that rhov <= 1
    x = x0 + N*T*np.power(1 - rhov, 3)

    if debug:
        with plt.Dp():
            fig, ax = plt.subplots()
            ax.set_xlabel("make_exclcurve_spline")
            ax.plot(x, y)
            ax.plot(trs, rgs, 'o', color='red')
            fig.tight_layout()
            plt.show()

    return UnivariateSpline(x, y, s=0, ext=3)


# base class for tk common dialogues
#
# this module provides a base class for accessing the common
# dialogues available in Tk 4.2 and newer.  use filedialog,
# colorchooser, and messagebox to access the individual
# dialogs.
#
# written by Fredrik Lundh, May 1997
#

"""
  mofified by Masatsuyo Takahashi, December 2015,
    1) make Dialog derive from Tk.Toplevel
    2) self.transient( self.parent )
    3) self.grab_set()
    4) self.withdraw()
    5) self.grab_release()
    6) changed self.master to self.parent
"""
from molass_legacy.KekLib.OurTkinter import Tk, TkVersion

class Dialog( Tk.Toplevel ):

    command  = None

    def __init__(self, master=None, **options):

        # FIXME: should this be placed on the module level instead?
        if TkVersion < 4.2:
            raise TclError("this module requires Tk 4.2 or newer")

        Tk.Toplevel.__init__( self )

        """
        changed self.master to self.parent, because
        self.master produced the following message during destruction.
          File "C:\Program Files\Python 3.5\lib\tkinter\__init__.py", line 2145, in destroy
        AttributeError: 'NoneType' object has no attribute 'children'
        """
        self.parent  = master
        self.options = options
        if not master and options.get('parent'):
            self.parent = options['parent']

        self.grab = False
        if 'grab' in options:
            self.grab = options.get('grab')
            del options[ 'grab' ]

        if self.parent and self.grab:
            self.transient( self.parent )
            self.grab_set()

        self.withdraw()

    def _fixoptions(self):
        pass # hook

    def _fixresult(self, widget, result):
        return result # hook

    def show(self, **options):

        # update instance options
        for k, v in options.items():
            self.options[k] = v

        self._fixoptions()

        # we need a dummy widget to properly process the options
        # (at least as long as we use Tkinter 1.63)
        w = Tk.Frame(self.master)

        try:

            s = w.tk.call(self.command, *w._options(self.options))

            s = self._fixresult(w, s)

        finally:

            try:
                # get rid of the widget
                w.destroy()
                if self.parent and self.grab:
                    self.grab_release()
            except:
                pass

        return s



"""
    Models.Stochastic.LnporeMoments.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import logging
import numpy as np
from scipy.optimize import minimize, basinhopping
from scipy.integrate import quad
from datetime import datetime
from importlib import reload
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy.Models.Stochastic.MomentUtils import compute_egh_moments, to_moderate_props
from molass_legacy.Models.Stochastic.LognormalUtils import compute_mu_sigma
import Models.Stochastic.LognormalPoreFunc
reload(Models.Stochastic.LognormalPoreFunc)
from molass_legacy.Models.Stochastic.LognormalPoreFunc import distr_func
from molass_legacy.Models.Stochastic.LnporeUtils import plot_lognormal_fitting_state
from molass_legacy.Models.Stochastic.MomentsStudy import monopore_study
from molass_legacy.Models.Stochastic.ParamLimits import MAX_PORESIZE, M1_WEIGHT, M2_WEIGHT

class LnporeMomentsEstimater:
    def __init__(self, lrf_src, parent=None, debug=False):
        self.logger = logging.getLogger(__name__)
        self.lrf_src = lrf_src
        self.parent = parent
        self.x = lrf_src.xr_x
        self.y = lrf_src.xr_y
        all_peaks = lrf_src.xr_peaks
        rgs, trs, orig_props, peak_rgs, peak_trs, props, indeces, qualities = lrf_src.compute_rgs(keep_num_components=False, debug=False)
        self.peaks = all_peaks[indeces]
        self.peak_rgs = peak_rgs
        self.props = props
        self.indeces = indeces

    def estimate(self, debug=False):
        import Models.Stochastic.MonoporeMoments
        reload(Models.Stochastic.MonoporeMoments)
        from molass_legacy.Models.Stochastic.MonoporeMoments import study_monopore_moments_impl

        num_peaks = len(self.peaks)
        egh_moments_list = compute_egh_moments(self.peaks)
        m_props = to_moderate_props(self.props)

        x = self.x
        y = self.y
        if False:
            mnp_params = monopore_study(x, y, peaks, peak_rgs, props, egh_moments_list, logger=logger, debug=False)
        else:
            mnp_params, temp_rgs, unreliable_indeces, params_scaler = study_monopore_moments_impl(self.lrf_src, debug=debug)
            self.logger.info("mnp_params=%s, temp_rgs=%s, unreliable_indeces=%s", mnp_params, temp_rgs, unreliable_indeces)
        N, T, x0, me, mp, poresize = mnp_params[0:6]

        nur = len(unreliable_indeces)

        use_basinhopping = False
        param_scales = np.array([100, 0.1, 50, 1, 0.1] + [10]*nur)  # scales to normalize the parameters for basinhopping

        def lnp_column_flex_objective(p, debug=False):
            if use_basinhopping:
                p *= param_scales
            N_, T_, x0_, mu, sigma = p[0:5]
            if nur > 0:
                temp_rgs[unreliable_indeces] = p[5:]

            dev_list = []
            for k, (M, rg) in enumerate(zip(egh_moments_list, temp_rgs)):
                M1_ = x0_ + N_ * T_   *  quad(lambda r : distr_func(r, mu, sigma) * (1 - min(1, rg/r))**(me +   mp), rg, MAX_PORESIZE)[0]
                M2_ =   2 * N_ * T_**2 * quad(lambda r : distr_func(r, mu, sigma) * (1 - min(1, rg/r))**(me + 2*mp), rg, MAX_PORESIZE)[0]
                dev_list.append(M1_WEIGHT*(M1_ - M[0])**2 + M2_WEIGHT*(M2_ - M[1])**2)
                if debug:
                    print([k], "M1, M1_ = %.3g, %.3g" % (M[0], M1_))
                    print([k], "M2, M2_ = %.3g, %.3g" % (M[1], M2_))
            return np.sum(np.asarray(dev_list)*m_props)

        stdev = poresize*0.2
        init_mu, init_sigma = compute_mu_sigma(poresize, stdev)
        init_params = [N, T, x0, init_mu, init_sigma] + list(temp_rgs[unreliable_indeces])
        eghM = egh_moments_list[0]
        x0_upper = eghM[0] - 5*np.sqrt(eghM[1])
        bounds = [(500, 20000), (0.01, 2), (-1000, x0_upper), (1, 10), (0.02, 1.0)] + [(10, 100)]*nur
        print("init_params=", init_params)
        print("bounds=", bounds)
        if use_basinhopping:
            init_params_ = np.array(init_params)/param_scales
            bounds_ = [(lower/scale, upper/scale) for (lower, upper), scale in zip(bounds, param_scales)]
            self.minima_counter = 0
            res = basinhopping(lnp_column_flex_objective, init_params_, niter=2,
                            callback=self.minima_callback,
                            minimizer_kwargs=dict(method='Nelder-Mead', bounds=bounds_))
            res.x *= param_scales   # do not forget to scale back the result
        else:
            res = minimize(lnp_column_flex_objective, init_params, bounds=bounds, method='Nelder-Mead')
        print("res.x=", res.x)

        N, T, x0, mu, sigma = res.x[0:5]
        if nur > 0:
            temp_rgs[unreliable_indeces] = res.x[5:]
        scales = self.props

        lnp_params = np.concatenate([(N, T, x0, me, mp, mu, sigma), scales])
        title = "lnp_column_objective result"
        ret = plot_lognormal_fitting_state(x, y, lnp_params, temp_rgs, title=title)

    def minima_callback(self, x, f, accept):
        self.minima_counter += 1
        print("%s: minima_counter=%d" % (datetime.now().strftime("%Y-%m-%d %H:%M:%S"), self.minima_counter))


"""
    SciPyMisc.py

    from https://stackoverflow.com/questions/22488637/getting-spline-equation-from-univariatespline-object

"""
import numpy as np
import matplotlib.pyplot as plt
from scipy.interpolate import splrep, UnivariateSpline, splev

def bspleval(x, knots, coeffs, order, debug=False):
    '''
    Evaluate a B-spline at a set of points.

    Parameters
    ----------
    x : list or ndarray
        The set of points at which to evaluate the spline.
    knots : list or ndarray
        The set of knots used to define the spline.
    coeffs : list of ndarray
        The set of spline coefficients.
    order : int
        The order of the spline.

    Returns
    -------
    y : ndarray
        The value of the spline at each point in x.
    '''

    k = order
    t = knots
    m = np.alen(t)
    npts = np.alen(x)
    B = np.zeros((m-1,k+1,npts))

    if debug:
        print('k=%i, m=%i, npts=%i' % (k, m, npts))
        print('t=', t)
        print('coeffs=', coeffs)

    ## Create the zero-order B-spline basis functions.
    for i in range(m-1):
        B[i,0,:] = np.float64(np.logical_and(x >= t[i], x < t[i+1]))

    if (k == 0):
        B[m-2,0,-1] = 1.0

    ## Next iteratively define the higher-order basis functions, working from lower order to higher.
    for j in range(1,k+1):
        for i in range(m-j-1):
            if (t[i+j] - t[i] == 0.0):
                first_term = 0.0
            else:
                first_term = ((x - t[i]) / (t[i+j] - t[i])) * B[i,j-1,:]

            if (t[i+j+1] - t[i+1] == 0.0):
                second_term = 0.0
            else:
                second_term = ((t[i+j+1] - x) / (t[i+j+1] - t[i+1])) * B[i+1,j-1,:]

            B[i,j,:] = first_term + second_term
        B[m-j-2,j,-1] = 1.0

    if debug:
        plt.figure()
        for i in range(m-1):
            plt.plot(x, B[i,k,:])
        plt.title('B-spline basis functions')

    ## Evaluate the spline by multiplying the coefficients with the highest-order basis functions.
    y = np.zeros(npts)
    for i in range(m-k-1):
        y += coeffs[i] * B[i,k,:]

    if debug:
        plt.figure()
        plt.plot(x, y)
        plt.title('spline curve')
        plt.show()

    return(y)

def bspleval_demo():
    x = np.array([-273.0, -176.4, -79.8, 16.9, 113.5, 210.1, 306.8, 403.4, 500.0])
    y = np.array([2.25927498e-53, 2.56028619e-03, 8.64512988e-01, 6.27456769e+00, 1.73894734e+01,
            3.29052124e+01, 5.14612316e+01, 7.20531200e+01, 9.40718450e+01])

    x_nodes = np.array([-273.0, -263.5, -234.8, -187.1, -120.3, -34.4, 70.6, 194.6, 337.8, 500.0])
    y_nodes = np.array([2.25927498e-53, 3.83520726e-46, 8.46685318e-11, 6.10568083e-04, 1.82380809e-01,
                    2.66344008e+00, 1.18164677e+01, 3.01811501e+01, 5.78812583e+01, 9.40718450e+01])

    ## Now get scipy's spline fit.
    k = 3
    tck = splrep(x_nodes, y_nodes, k=k, s=0)
    knots = tck[0]
    coeffs = tck[1]
    print('knot points=', knots)
    print('coefficients=', coeffs)

    ## Now try scipy's object-oriented version. The result is exactly the same as "tck": the knots are the
    ## same and the coeffs are the same, they are just queried in a different way.
    uspline = UnivariateSpline(x_nodes, y_nodes, s=0)
    uspline_knots = uspline.get_knots()
    uspline_coeffs = uspline.get_coeffs()

    ## Here are scipy's native spline evaluation methods. Again, "ytck" and "y_uspline" are exactly equal.
    ytck = splev(x, tck)
    y_uspline = uspline(x)
    y_knots = uspline(knots)

    ## Now let's try our manually-calculated evaluation function.
    y_eval = bspleval(x, knots, coeffs, k, debug=False)

    plt.plot(x, ytck, label='tck')
    plt.plot(x, y_uspline, label='uspline')
    plt.plot(x, y_eval, label='manual')

    ## Next plot the knots and nodes.
    plt.plot(x_nodes, y_nodes, 'ko', markersize=7, label='input nodes')            ## nodes
    plt.plot(knots, y_knots, 'mo', markersize=5, label='tck knots')    ## knots
    plt.xlim((-300.0,530.0))
    plt.legend(loc='best', prop={'size':14})

    plt.figure()
    plt.title('difference')
    plt.plot(x, ytck-y_uspline, label='tck-uspl')
    plt.plot(x, ytck-y_eval, label='tck-manual')
    plt.legend(loc='best', prop={'size':14})

    plt.show()



"""

    FileInfoTable.py

    Copyright (c) 2016-2023, SAXS Team, KEK-PF

"""

import os
import glob
import re
import time
from datetime               import datetime as dt
from molass_legacy.KekLib.OurTkinter             import Tk, ToolTip, ttk
try:
    import molass_legacy.KekLib.CustomMessageBox as MessageBox
except:
    import OurMessageBox as MessageBox
from TkMiniTable            import TkMiniTable
from molass_legacy._MOLASS.SerialSettings         import get_setting
from ResultGui              import ResultGui
from KekToolsGP             import autorg   as autorg_kekpf
from molass_legacy.KekLib.NumpyUtils             import np_loadtxt
from SerialDataUtils        import get_xray_files

class FileInfoTable( Tk.Frame ):
    def __init__( self, parent, dialog, loader, fixed_font, *args, **kw ):

        Tk.Frame.__init__( self, parent, *args, **kw )

        self.parent = parent
        self.dialog = dialog
        self.high_dpi = dialog.high_dpi
        self.loader = loader
        self.fixed_font = fixed_font
        self.create_table()
        self.datafiles  = None
        # self.refresh_impl()

        self.popup_menu_ready = False

        parent.update()

    def create_table( self ):
        self.averaged_intensity_array = None

        self.table = TkMiniTable( self, columns=[ '_', 'filename', 'datetime', 'checkbutton' ] )
        self.table.pack( fill=Tk.BOTH, expand=1 )

        self.table.add_mousewheel_bind( self.dialog )
        # it seemd that '<Control-c>' should be bound from the top-level.
        self.dialog.bind( '<Control-c>', self.table._on_control_c )

    def create_popup_menu( self, destroy=False ):
        if destroy:
            if self.popup_menu_ready:
                self.menu.destroy()

        atsas_exe_paths = get_setting( 'atsas_exe_paths' )
        atsas_versions = self.get_atsas_versions( atsas_exe_paths )

        # create a popup menu
        self.menu = Tk.Menu( self, tearoff=0 )
        num_atsas_exe = len(atsas_exe_paths)
        if num_atsas_exe > 0:
            for i in range( num_atsas_exe ):
                self.menu.add_command( label='Run with ' + atsas_versions[i],  command=lambda i_ = i: self.run_autorg( i_ ) )
        else:
            self.menu.add_command( label='Run AutoRg',  command=lambda: self.run_autorg() )

        self.menu.add_command( label='Run AutoGuinier Animation',  command=lambda: self.run_simple_guinier_animation() )

    def get_atsas_versions( self, atsas_exe_paths ):
        from molass_legacy.ATSAS.AtsasUtils import get_versions
        return ['ATSAS ' + v for v in get_versions()]

    def _on_right_click( self, event ):
        if not self.popup_menu_ready:
            self.create_popup_menu()
            self.popup_menu_ready = True

        self.popup_position = ( event.x, event.y )
        # print( 'popup_position=', self.popup_position )

        self.menu.post( event.x_root, event.y_root )

    def refresh(self):
        self.table.destroy()
        self.create_table()
        self.refresh_impl()

    def refresh_impl(self):

        self.testing = self.dialog.testing
        table = self.table

        table.heading( "_", text='No' )
        table.column(  "_", width=4, stretch=0  )
        table.heading( 'filename', text='File Name')
        width = 53 if self.high_dpi else 42
        table.column(  'filename', anchor='center', width=width )
        table.heading( 'datetime', text='Last Modified')
        width = 20 if self.high_dpi else 16
        table.column(  'datetime', anchor='center', width=width, stretch=0 )
        table.heading( 'checkbutton', text='exclude')
        table.column(  'checkbutton', anchor='center', width=7, stretch=0 )

        self.in_folder   = get_setting( 'in_folder' )
        self.datafiles  = get_xray_files( self.in_folder )

        n = 0
        if self.in_folder is not None:
            for path in self.datafiles:
                t   = os.path.getmtime( path )
                dt_ = dt.fromtimestamp( t )
                dir_, file = os.path.split( path )
                table.insert_row( ( n, file, dt_.strftime( '%Y-%m-%d %H:%M:%S' ), '' ) )
                n += 1
        self.num_rows = n

    def _on_click( self, event ):
        # print( '_on_click' )
        cell = self.table.active_cell
        if cell is None: return

        if cell.row < 1 or cell.row > self.num_rows:
                return

        self.dialog.draw_figure(selected=cell.row-1)

        if get_setting('data_exclusion'):
            if cell.col == 3:
                # print( 'cell.row=',  cell.row, 'cell.col=', cell.col )
                if cell.svar.get() == '':
                    if not self.check_too_many_excluded_neighbor_cells( cell.row ):
                        cell.svar.set( 'X' )
                else:
                    cell.svar.set( '' )

    def check_too_many_excluded_neighbor_cells( self, row ):
        return False

    def _on_double_click( self, event ):
        self.run_autorg( 0 )

    def run_autorg( self, exe_index=None ):
        self.config( cursor='wait' )
        self.update()

        data, file, _, n = self.get_data_array()

        print( 'run_autorg: data.shape=', data.shape )

        try:
            q_limit_index  = self.serial_data.get_usable_q_limit()
        except:
            from molass_legacy.KekLib.ExceptionTracebacker   import ExceptionTracebacker
            etb = ExceptionTracebacker()
            print(etb)
            q_limit_index  = None

        orig_info = [ self.serial_data, n ]
        result_gui = ResultGui( self, self.in_folder, file, np_array=data, atsas_exe_index=exe_index, orig_info=orig_info, q_limit_index=q_limit_index )

        self.config( cursor='' )
        self.update()

        result_gui.after( 0, result_gui.focus_force )       # workaround to raise to forefront

    def get_data_array( self, n=None ):

        if n is None:
            row = self.table.selected_rows[0]
            # file = self.table.cells_array[ row, 1 ].svar.get()
            n = row -1

        file = self.table.cells_array[ n+1, 1 ].svar.get()

        if self.loader.is_loadable:
            serial_data = self.loader.get_data_object()
            usable_slice = serial_data.get_usable_slice()
            if self.averaged_intensity_array is None:
                num_curves_averaged = get_setting( 'num_curves_averaged' )
                self.averaged_intensity_array, _, _ = serial_data.get_averaged_data( num_curves_averaged )
            data = self.averaged_intensity_array[n, usable_slice]
            title_averaged = ' (averaged)'
            self.serial_data = serial_data
        else:
            data, _ = np_loadtxt( self.in_folder + '/' + file )
            title_averaged = ''
            self.serial_data = None

        # print( 'file[%d]=%s%s' % ( n, file, title_averaged ) )

        return data, file, title_averaged, n

    def run_simple_guinier_animation( self ):
        self.config( cursor='wait' )
        self.update()

        from SimpleGuinier          import SimpleGuinier
        from molass_legacy.AutorgKekAdapter       import AutorgKekAdapter
        from SimpleGuinierAnimation import SimpleGuinierAnimation
        from CanvasDialog           import CanvasDialog

        data, file, title_averaged, n = self.get_data_array()

        guinier = SimpleGuinier( data, anim_data=True )
        adapter = AutorgKekAdapter( None, guinier=guinier )
        adapter_result = adapter.run()

        self.anim_dialog = CanvasDialog( "AutoGuinier Animation", parent=self.parent )
        anim_iter_max = 300

        def anim_func( fig  ):
            anim = SimpleGuinierAnimation( guinier, title=file+title_averaged, result_rg=adapter_result.Rg  )
            anim.draw( fig, anim_iter_max=anim_iter_max )
            self.anim_dialog.mpl_canvas.draw()

        self.anim_dialog.show( anim_func, figsize=( 18, 5 ) )

        self.config( cursor='' )
        self.update()

    def get_excluded_indeces( self ):
        indeces = []
        for i, row in enumerate( self.table.cells_array, start=-1 ):
            if row[3].svar.get() == 'X':
                indeces.append( i )
        return indeces

    def set_exclude( self, exclude ):
        for i in exclude:
            self.table.cells_array[ i+1, 3 ].svar.set( 'X' )

    def select_row(self, row):
        if row < self.num_rows:
            self.table.select_row(row)



"""

    GuiParts.UvFolderEntry.py

    Copyright (c) 2016-2024, SAXS Team, KEK-PF

"""
import os
from molass_legacy.KekLib.OurTkinter import is_empty_val
from molass_legacy.KekLib.TkCustomWidgets import FolderEntry
import molass_legacy.KekLib.CustomMessageBox as MessageBox
from molass_legacy._MOLASS.SerialSettings import get_setting

class UvFolderEntry(FolderEntry):
    def check( self, *args ):
        if get_setting( 'use_xray_conc' ) == 1:
            return True

        f = self.variable.get()
        ret = False
        if is_empty_val( f ):
            self.set_error()
            message = (   'This folder input is required.\n'
                        + 'Enable "use xray-proportional concentration" in the "Specialist Options" '
                        + 'if you would like to analyze without UV-data.'
                        )
            MessageBox.showerror( "Folder Input Error", message, parent=self.parent )
        elif not os.path.exists( f ):
            self.set_error()
            MessageBox.showerror( "Folder Input Error", "'%s' does not exist." % f, parent=self.parent )
        elif not os.path.isdir( f ):
            self.set_error()
            MessageBox.showerror( "Folder Input Error", "'%s' is not a folder." % f, parent=self.parent )
        else:
            self.entry.config( fg='black' )
            ret = True

        if not ret:
            self.entry.focus_force()

        return ret



"""
    MwIntegrity.py

    Copyright (c) 2023-2025, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.spatial import distance
import matplotlib.pyplot as plt
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from molass_legacy.KekLib.OurTkinter import Tk, Dialog
from molass_legacy.KekLib.OurMatplotlib import NavigationToolbar, get_color
from molass_legacy._MOLASS.SerialSettings import get_setting
from molass_legacy.SerialAnalyzer.DataUtils import get_in_folder
# from molass_legacy.DENSS.DenssUtils import fit_data_impl
from molass_legacy.GuinierAnalyzer.SimpleGuinier import SimpleGuinier
from .LrfExporter import LrfExporter
from .FvScoreConverter import convert_score

BASE_WEIGHT = 0.9
PLUS_WEIGHT = 0.1

def compute_mw_integrity_impl(ratios, integer_ratios, preceder):
    integrity = np.log10(distance.jensenshannon(ratios, integer_ratios))
    if integrity < preceder:
        score = preceder*BASE_WEIGHT + integrity*PLUS_WEIGHT
    else:
        score = integrity
    return score

class MwIntegrityPlot(Dialog):
    def __init__(self, parent, js_canvas):
        self.parent = parent
        self.js_canvas = js_canvas
        self.class_code = js_canvas.dialog.class_code
        self.elution_model = js_canvas.elution_model
        self.fullopt = js_canvas.fullopt
        self.params = js_canvas.get_current_params()
        self.separate_params = self.fullopt.split_params_simple(self.params)
        exporter = LrfExporter(self.fullopt, self.params, js_canvas.dsets)
        self.info = exporter.xr_result.info
        self.integer_ratios = get_setting("mw_integer_ratios")
        Dialog.__init__(self, parent, "MW Integer Ratios", visible=False)

    def show(self):
        self._show()

    def body(self, body_frame):
        cframe = Tk.Frame(body_frame)
        cframe.pack()
        tframe = Tk.Frame(body_frame)
        tframe.pack(fill=Tk.X)

        fig, axes = plt.subplots(ncols=3, figsize=(18,5))

        self.draw_mws(fig, axes)

        fig.tight_layout()
        self.fig = fig
        self.mpl_canvas = FigureCanvasTkAgg(self.fig, cframe)
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack(fill=Tk.BOTH, expand=1)
        self.toolbar = NavigationToolbar(self.mpl_canvas, tframe)
        self.toolbar.update()

    def buttonbox(self):
        box = Tk.Frame(self)
        box.pack()
        w = Tk.Button(box, text="Close", width=10, command=self.cancel)
        w.pack(side=Tk.LEFT, padx=5, pady=5)

        self.bind("<Return>", self.ok)
        self.bind("<Escape>", self.cancel)

    def draw_mws(self, fig, axes):
        from importlib import reload
        import Optimizer.IntegerRatios as module
        reload(module)
        from .IntegerRatios import determine_integer_ratios

        M, M_, E, P, C, qv = self.info

        ax1, ax2, ax3 = axes

        X = np.linalg.pinv(M) @ P           # M or M_
        Ep = np.sqrt((E**2) @ (X**2))

        in_folder = get_in_folder()
        fig.suptitle("Molecular Weight Integer Ratios Plot on %s with func=%s" % (in_folder, self.class_code), fontsize=20)

        ax1.set_title("Component Elutions in UV", fontsize=16)

        axes_ = (ax1, None, None, None)
        self.fullopt.objective_func(self.params, plot=True, axis_info=(fig, axes_))

        ax2.semilogy()
        ax2.set_title( "Component Scattering Profiles", fontsize=16)
        ax2.set_xlabel('Q')
        ax2.set_ylabel('Intensity ($Log_{10}$)')

        mw_list = []
        for k in range(P.shape[1] - 1):
            a = P[:,k]
            e = Ep[:,k]
            c = np.max(C[k,:])

            data = np.array([qv, a, e]).T
            sg = SimpleGuinier(data)
            mw = sg.Iz
            print([k], "Iz=%g" % sg.Iz, "Iz/C=%g" % mw)
            mw_list.append(mw)

            ax2.plot(qv, a, label='component-%d' % (k+1))

        ax2.legend()

        mw_array = np.array(mw_list)
        nc = len(mw_array)
        if self.integer_ratios is None:
            self.integer_ratios = determine_integer_ratios(mw_array)
        dist = compute_mw_integrity_impl(mw_array, self.integer_ratios, -10)
        score = convert_score(dist)
        names = ["component-%d" % (k+1) for k in range(nc)]
        colors = ["C%d" % k for k in range(nc)]

        ax3.set_title("Integrity Score: %.2g compared to %s" % (score, str(self.integer_ratios)), fontsize=16)
        ax3.set_ylabel("Molecular Weight: I(0)/C")
        ax3.bar(names, mw_array, color=colors)



# coding: utf-8
"""
    UnfoldingTutorial.py

    Copyright (c) 2019, SAXS Team, KEK-PF
"""
import os
import numpy as np
from lmfit import Parameters, minimize
import matplotlib.pyplot as plt
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from molass_legacy.KekLib.OurTkinter import Tk, Dialog, ttk, is_empty_val
from molass_legacy.KekLib.OurMatplotlib import NavigationToolbar
from molass_legacy.KekLib.TkSupplements import tk_set_icon_portable
from molass_legacy.KekLib.TkCustomWidgets import FolderEntry
from molass_legacy.KekLib.TkUtils import is_low_resolution
from MenuButton import MenuButton
from UnfoldingModel import UnfoldingModel

class UnfoldingTutorial(Dialog):
    def __init__(self, parent):
        self.busy = False
        self.applied = False
        self.data_list = [None]*3
        self.zlim_3d = None
        self.ylim = None
        Dialog.__init__(self, parent, "Unfolding Model Tutorial", visible=False)

    def body(self, body_frame):
        tk_set_icon_portable(self)

        iframe = Tk.Frame(body_frame)
        iframe.pack(padx=20, pady=10)
        cframe = Tk.Frame(body_frame)
        cframe.pack()
        tbframe = Tk.Frame(body_frame)
        tbframe.pack(fill=Tk.X, expand=1)

        self.build_entryframe(iframe)
        self.build_canvas(cframe, tbframe)
        self.solve()

    def build_entryframe(self, iframe):
        table_frame = Tk.Frame(iframe)
        table_frame.pack()
        button_frame = Tk.Frame(iframe)
        button_frame.pack(fill=Tk.X)

        label = Tk.Label(table_frame, text="Given")
        label.grid(row=0, column=1)
        label = Tk.Label(table_frame, text="Solved")
        label.grid(row=0, column=2)

        self.texts = texts = ["G", "m", "af", "bf", "au", "bu"]
        values = [4, 1, 0.5, -0.01, 0.3, -0.02]

        self.vars1 = []
        self.vars2 = []

        for i, t in enumerate(texts):
            k = i+1
            label = Tk.Label(table_frame, text=t)
            label.grid(row=k, column=0, sticky=Tk.E)
            var = Tk.DoubleVar()
            var.set(values[i])
            self.vars1.append(var)
            entry = Tk.Entry(table_frame, textvariable=var, justify=Tk.RIGHT, width=8)
            entry.grid(row=k, column=1, sticky=Tk.W, padx=10)
            var = Tk.DoubleVar()
            self.vars2.append(var)
            label = Tk.Label(table_frame, textvariable=var, justify=Tk.RIGHT, width=8, bg='white')
            label.grid(row=k, column=2)

        ex_menu = MenuButton(button_frame, "Examples", [
                            ("Example1(Regular)", self.solve_regular),
                            ("Example2(Irregular)", self.solve_irregular),
                            ])
        ex_menu.pack(side=Tk.LEFT)

        button = Tk.Button(button_frame, text="Solve", command=self.solve)
        button.pack(side=Tk.RIGHT)

    def build_canvas(self, cframe, tbframe):
        figsize = (17,5) if is_low_resolution() else (23,7)
        fig, axes = plt.subplots(nrows=1, ncols=3, figsize=figsize)
        self.fig = fig
        self.axes = axes
        self.axest = [ax.twinx() for ax in axes]
        self.mpl_canvas = FigureCanvasTkAgg( self.fig, cframe )
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack(fill=Tk.BOTH, expand=1)
        self.mpl_canvas.mpl_connect( 'button_press_event', self.on_button_press )
        self.toolbar = NavigationToolbar(self.mpl_canvas, tbframe)
        self.toolbar.update()
        fig.subplots_adjust(top=0.92, bottom=0.1, left=0.05, right=0.95, wspace=0.2)

    def show(self):
        self._show(wait=False)
        self.waited_widget = Tk.Frame(self)
        self.wait_window(self.waited_widget)

    def on_button_press(self, event):
        if event.dblclick:
            self.toggle3d()

    def ok(self):
        self.apply()
        self.close()

    def exit(self):
        self.close()

    def close(self):
        self.withdraw()
        self.update_idletasks()
        self.grab_release()
        self.waited_widget.destroy()

    def apply(self):
        self.applied = True

    def make_params(self):
        params = Parameters()
        for k, var in enumerate(self.vars1):
            t = self.texts[k]
            params.add(t, var.get())
        return params

    def draw_ab_params(self, ax, axt):
        ax.cla()
        axt.cla()
        axt.set_axis_off()
        ax.set_title("Given Params(a*,b*)", fontsize=24)
        ax.set_xlabel('a')
        ax.set_ylabel('b')

        xmax = None
        ymax = None
        for t, i, j in [("(af,bf)",2,3), ("(au,bu)",4,5)]:
            x = self.vars1[i].get()
            y = self.vars1[j].get()
            if xmax is None or abs(x) > xmax:
                xmax = abs(x)
            if ymax is None or abs(y) > ymax:
                ymax = abs(y)
            ax.text(x, y, t)
            ax.annotate('', xy=(x, y), xytext=(0,0),
                arrowprops=dict(headwidth=5, width=0.5, color='black', shrink=0)
               )
        xmax_ = max(1, xmax)
        ymax_ = max(0.03, ymax)
        ax.set_xlim(-xmax_, xmax_)
        ax.set_ylim(-ymax_, ymax_)

    def draw_test_data(self, ax, axt):
        x = np.linspace(0, 8, 100)

        model = UnfoldingModel()
        params = self.make_params()
        Pf = model.compute_Pf(params, x)
        yf = model.compute_yf(params, Pf, x)
        yu = model.compute_yu(params, Pf, x)
        yfyu = yf+yu
        self.draw(ax, axt, "Given Data", x, Pf, yf, yu, yfyu)
        return yfyu

    def draw(self, ax, axt, title, x, Pf, yf, yu, yfyu):
        ax.cla()
        axt.cla()
        ax.set_title(title, fontsize=24)
        ax.set_xlabel('[Denaturant]')
        ax.set_ylabel('Proportion')
        axt.set_ylabel('Intensity')
        axt.grid(False)
        ax.plot(x, Pf, label=r'$pf=\frac{1}{1+exp(-\frac{(G-m*x)}{RT})}$')
        axt.plot(x, yf, ':', label='yf')
        axt.plot(x, yu, ':', label='yu')
        axt.plot(x, yfyu, label='yf+yu')
        ax.legend(loc='upper center', fontsize=16)
        axt.legend(loc='center right', fontsize=16)

    def draw_solved_data(self, ax, axt, data):
        model = UnfoldingModel()
        x = np.linspace(0, 8, 100)
        y = data
        res = model.fit(x, y)
        params = res.params
        # print('params=', params)
        Pf = model.compute_Pf(params, x)
        yf = model.compute_yf(params, Pf, x)
        yu = model.compute_yu(params, Pf, x)
        yfyu = yf+yu
        self.draw(ax, axt, "Solved Data", x, Pf, yf, yu, yfyu)

        for k, var in enumerate(self.vars2):
            t = self.texts[k]
            f = float('%g' % params[t].value)
            var.set(f)

    def solve(self):
        self.draw_ab_params(self.axes[0], self.axest[0])
        yfyu = self.draw_test_data(self.axes[1], self.axest[1])
        self.draw_solved_data(self.axes[2], self.axest[2], yfyu)
        self.mpl_canvas.draw()

    def set_params(self, values):
        for w, v in zip(self.vars1, values):
            w.set(v)

    def solve_regular(self):
        self.set_params([4, 1, 0.5, -0.01, 0.3, -0.02])
        self.solve()

    def solve_irregular(self):
        self.set_params([4, 1, 0.5, -0.01, 0.5, -0.01])
        self.solve()



# coding: utf-8
"""
    DecompDemo.py.py

    Copyright (c) 2018, Masatsuyo Takahashi, KEK-PF
"""
import copy
import numpy                as np
import matplotlib.pyplot    as plt
import matplotlib.gridspec  as gridspec
import matplotlib.patches   as mpl_patches      # 'as patches' does not work properly
from matplotlib.backends.backend_tkagg  import FigureCanvasTkAgg
from molass_legacy.KekLib.OurTkinter             import Tk, Dialog
from molass_legacy.KekLib.TkUtils                import is_low_resolution
from molass_legacy.KekLib.TkSupplements          import tk_set_icon_portable
from molass_legacy.KekLib.OurMatplotlib          import NavigationToolbar, get_default_colors
from molass_legacy.Models.ElutionCurveModels     import EMG
from XrayDecomposer         import XrayDecomposer

class DecompDemo( Dialog ):
    def __init__( self, parent, title, sd, mapper, peaks=None ):
        self.parent = parent
        self.title_ = title
        self.sd     = sd
        self.mapper = mapper
        self.peaks  = peaks

    def show( self ):
        Dialog.__init__( self, self.parent, self.title_ )

    def body( self, body_frame ):   # overrides parent class method
        tk_set_icon_portable( self )

        cframe = Tk.Frame( body_frame )
        cframe.pack()

        emg_model = EMG()

        gs = gridspec.GridSpec( 4, 2 )
        figsize = ( 15, 8 ) if is_low_resolution() else ( 18, 10 )
        fig = plt.figure( figsize=figsize )
        ax2_ = fig.add_subplot( gs[0:3,0] )
        ax3_ = fig.add_subplot( gs[0:3,1] )
        ax5_ = fig.add_subplot( gs[3,0] )
        ax6_ = fig.add_subplot( gs[3,1] )
        axes_list   = [ [ax2_, ax3_], [ax5_, ax6_] ]
        name_list   = [ "EGH" ]
        model_list  = [ None ]

        self.fig = fig
        self.mpl_canvas = FigureCanvasTkAgg( self.fig, cframe )
        self.mpl_canvas.draw()
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack( fill=Tk.BOTH, expand=1 )

        self.draw( axes_list, name_list, model_list )

        self.fig.tight_layout()
        self.mpl_canvas.draw()

        self.toolbar = NavigationToolbar( self.mpl_canvas, cframe )
        self.toolbar.update()

    def buttonbox(self):
        '''add standard button box.

        override if you do not want the standard buttons
        '''

        box = Tk.Frame(self)

        w = Tk.Button(box, text="OK", width=10, command=self.ok, default=Tk.ACTIVE)
        w.pack(side=Tk.LEFT, padx=5, pady=5)
        self.ok_button = w
        w = Tk.Button(box, text="Cancel", width=10, command=self.cancel)
        w.pack(side=Tk.LEFT, padx=5, pady=5)
        self.cancel_button = w

        self.bind("<Return>", self.ok)
        self.bind("<Escape>", self.cancel)

        box.pack()

    def draw( self, axes_list, name_list, model_list ):
        mapper  = self.mapper
        sd      = self.sd

        x_curve = mapper.x_curve
        a_baseline = mapper.a_base + mapper.a_base_adjustment
        x_baseline = mapper.x_base + mapper.x_base_adjustment

        x_corrected = mapper.x_vector - x_baseline
        sd.apply_baseline_correction( mapper.get_mapped_info() )
        data    = sd.intensity_array

        y   = x_corrected
        x   = np.arange( len(y) )

        if self.peaks is None:
            self.peaks = peaks = x_curve.get_emg_peaks()
        else:
            peaks   = self.peaks

        print( 'peaks=', peaks )

        self.chisqr_info_list = []

        k = 0
        ax2, ax3 = axes_list[k]
        ax5, ax6 = axes_list[1]
        name = name_list[k]
        fa1  = XrayDecomposer( x_curve, x, y, data, peaks=None, retry_valley=True, model=None, deeply=True )
        fa2  = XrayDecomposer( x_curve, x, y, data, peaks=peaks, retry_valley=True, model=None, deeply=True )

        ax2.set_title( "Decomposition as two peaks", fontsize=16 )
        ax3.set_title( "Decomposition as three peaks", fontsize=16 )

        ax5.set_title( "Residuals", fontsize=16 )
        ax6.set_title( "Residuals", fontsize=16 )

        ax2.plot( x_corrected, color='orange' )
        ax3.plot( x_corrected, color='orange' )

        default_colors = get_default_colors()

        def draw_decomposition( ax, ax_, fit_recs, colors ):

            x_residual = copy.deepcopy( x_corrected )

            for i, rec in enumerate( fit_recs ):
                func    = rec[1]
                ey = func( x )
                ax.plot( ey, linewidth=5.0, color=colors[i] )
                x_residual -= ey

            ax_.plot( x_residual, color='gray' )

        draw_decomposition( ax2, ax5, fa1.init_fit_recs, [ default_colors[0], default_colors[1] ] )
        draw_decomposition( ax3, ax6, fa2.fit_recs, [ default_colors[0], default_colors[2], default_colors[1] ] )

        ax6.set_ylim( ax5.get_ylim() )

        for ax in [ ax2, ax3, ax5, ax6 ]:
            ax.tick_params( labelsize=16 )

        if False:

            for ax in axes_list[k]:
                ymin, ymax = ax.get_ylim()
                for peak in peaks:
                    f = peak.flimL
                    t = peak.flimR
                    p = mpl_patches.Rectangle(
                            (f, ymin),      # (x,y)
                            t - f,          # width
                            ymax - ymin,    # height
                            facecolor   = 'cyan',
                            alpha       = 0.2,
                        )
                    ax.add_patch( p )



"""
    EdmParams.py

    Copyright (c) 2023-2025, SAXS Team, KEK-PF
"""
import logging
import numpy as np
from .BaselineParams import get_num_baseparams
from molass_legacy._MOLASS.SerialSettings import get_setting
from molass_legacy.Models.RateTheory.EDM import edm_impl, MIN_CINJ, MAX_CINJ
from molass_legacy.Optimizer.BasicOptimizer import AVOID_VANISHING_RATIO

NUM_COL_PARAMS = 1
NUM_ELEMENT_PARAMS = 7

def get_common_parameter_names(nc):
    xr_names = []
    for k in range(nc):
        xr_names += ["$t0_%d$" % k, "$u_%d$" % k, "$a_%d$" % k, r"$\b_%d$" % k, r"$\e_%d$" % k, r"$\Dz_%d$" % k, r"cinj_%d" % k]

    rg_names = ["$R_{g%d}$" % k for k in range(nc)]
    mapping_names = ["$mp_a$", "$mp_b$"]
    uv_names = ["$uh_%d$" % k for k in range(nc)]
    mr_names = ["$mr_a$", "$mr_b$"]
    edmcol_names = ["$Tz$"]

    return xr_names, rg_names, mapping_names, uv_names, mr_names, edmcol_names

class EdmParams:
    def __init__(self, n_components):
        self.logger = logging.getLogger(__name__)
        self.n_components = n_components
        self.num_baseparams = get_num_baseparams()
        self.integral_baseline = self.num_baseparams == 3
        self.t0_upper_bound = get_setting("t0_upper_bound")
        self.use_K = False

        nc = n_components - 1

        self.pos = []
        self.pos.append(0)      # [0] xr_params
        sep = nc*NUM_ELEMENT_PARAMS
        self.pos.append(sep)    # [1] xr_baseparams
        sep += self.num_baseparams
        self.pos.append(sep)    # [2] rgs
        sep += nc
        self.pos.append(sep)    # [3] mapping
        sep = sep+2
        self.pos.append(sep)    # [4] uv_params
        sep += nc
        self.pos.append(sep)    # [5] uv_baseparams
        sep += 5 + self.num_baseparams
        self.pos.append(sep)    # [6] mappable_range
        sep += 2
        self.pos.append(sep)    # [7] end

        self.num_params = sep   # note that this does not include sec params

        self.logger.info("pos=%s", str(self.pos))

    def __str__(self):
        return "EdmParams(nc=%d)" % (self.n_components)

    def get_model_name(self):
        return 'EDM'

    def set_x(self, x):
        self.x = x

    def get_estimator(self, editor, debug=True):
        if debug:
            from importlib import reload
            import molass_legacy.Estimators.EdmEstimator
            reload(molass_legacy.Estimators.EdmEstimator)
        from molass_legacy.Estimators.EdmEstimator import EdmEstimator

        estimator = EdmEstimator(editor, self.n_components)
        self.t0_upper_bound = estimator.get_t0_upper_bound()
        return estimator

    def compute_init_guess(self, *args):
        assert False

    def compute_comformance(self, *args, **kwargs):
        assert False

    def estimate_conformance_params(self, *args):
        assert False

    def get_xr_param_bounds(self, xr_params):

        xr_h_max = np.max(xr_params[:,-1])      # cinj
        xr_h_min = xr_h_max*AVOID_VANISHING_RATIO
        xr_bounds = []
        for t0, u, a, b, e, Dz, cinj in xr_params:
            xr_bounds.append((-500, 1000))      # t0
            xr_bounds.append((0.00001, 50.0))   # u
            xr_bounds.append((0.0001, 1.0))     # a
            xr_bounds.append((-20.0, +20.0))    # b
            xr_bounds.append((0.001,  1.0))     # e
            xr_bounds.append((0.001, 40.0))     # Dz
            xr_bounds.append((max(MIN_CINJ, xr_h_min), min(MAX_CINJ, xr_h_max*2)))      # cinj

        return xr_bounds

    def get_param_bounds(self, params):

        init_xr_params, init_xr_baseparams, init_rgs, init_mapping, init_uv_params, init_uv_baseparams, init_mappable_range = self.split_params_simple(params)[0:7]

        m_allow = 100

        xr_bounds = self.get_xr_param_bounds(init_xr_params)

        for k, v in enumerate(init_xr_baseparams):
            v_allow = max(0.1, abs(v))*0.2          # note that v may be zero
            if self.integral_baseline and k == 2:
                bounds = (max(0, v - v_allow), v + v_allow)
            else:
                bounds = (v - v_allow, v + v_allow)
            xr_bounds.append(bounds)

        rg_bounds = [(rg*0.5, rg*2) for rg in init_rgs]
        a, b = init_mapping
        mapping_bounds = [(a*0.8, a*1.2), (-m_allow, m_allow)]

        uv_h_max = np.max(init_uv_params)
        uv_h_min = uv_h_max*AVOID_VANISHING_RATIO
        uv_bounds = [(uv_h_min, uv_h_max*2) for h in init_uv_params]
        for k, v in enumerate(init_uv_baseparams):
            v_allow = max(0.1, abs(v))*0.2          # note that v may be zero
            if self.integral_baseline and k == 7:
                bounds = (max(0, v - v_allow), v + v_allow)
            else:
                bounds = (v - v_allow, v + v_allow)
            uv_bounds.append(bounds)

        f, t = init_mappable_range
        dx = (t - f)*0.1
        range_bounds = [(f-dx, f+dx), (t-dx, t+dx)]

        # u
        edm_colparams = [(0.45, 0.55)]

        self.bounds_lengths = [len(b) for b in [xr_bounds, rg_bounds, mapping_bounds, uv_bounds, range_bounds, edm_colparams]]
        self.logger.info("bounds_lengths=%s", str(self.bounds_lengths))
        return xr_bounds + rg_bounds + mapping_bounds + uv_bounds + range_bounds + edm_colparams

    def split_bounds(self, bounds):
        separate_bounds = []
        offset = 0
        for k, length in enumerate(self.bounds_lengths):
            next_offset = offset + length
            sep_bounds = bounds[offset:next_offset]
            if k == 0:
                sep_bounds = self.reshape_xr_bounds(sep_bounds)
            separate_bounds.append(sep_bounds)
            offset = next_offset
        return separate_bounds

    def update_bounds_hook(self, masked_init_params):
        # nothing to do
        pass

    def get_peak_pos_array_list(self, x_array):
        nc = self.n_components - 1
        gr_start = self.get_rg_start_index()
        x = self.x

        pos_array_list = []
        for params in x_array:

            # these must be unified with split_params_simple
            xr_params = params[0:nc*NUM_ELEMENT_PARAMS].reshape((nc,NUM_ELEMENT_PARAMS))

            pos = []
            for t0, u, a, b, e, Dz, cinj in xr_params:
                xr_cy = edm_impl(x, t0, u, a, b, e, Dz, cinj)
                j = np.argmax(xr_cy)
                pos.append(x[j])
            pos_array_list.append(pos)

        return np.array(pos_array_list).T

    def get_params_sheet(self, parent, params, dsets, optimizer, debug=True):
        if debug:
            from importlib import reload
            import molass_legacy.ModelParams.EdmParamsSheet
            reload(molass_legacy.ModelParams.EdmParamsSheet)
        from .EdmParamsSheet import EdmParamsSheet
        return EdmParamsSheet(parent, params, dsets, optimizer)

    def get_adjuster(self, debug=True):
        if debug:
            from importlib import reload
            import molass_legacy.ModelParams.StcAdjuster
            reload(molass_legacy.ModelParams.StcAdjuster)
        from .StcAdjuster import StcAdjuster
        return StcAdjuster()

    def split_params(self, n, params):
        if self.num_params != len(params):
            raise ValueError("len(params)=%d != %d, which is calculated from number of components %d." % (len(params), self.num_params, n))

        nc = self.n_components - 1
        ret_params = []
        k = 0
        for p, q in zip(self.pos[:-1], self.pos[1:]):
            if k == 0:
                params_ = params[p:q].reshape((nc, NUM_ELEMENT_PARAMS))
            else:
                params_ = params[p:q]
            ret_params.append(params_)
            k += 1

        return ret_params

    def split_params_simple(self, params):
        decomp_params = params[:-NUM_COL_PARAMS]
        seccol_params = params[-NUM_COL_PARAMS:]
        self.separate_params = self.split_params(self.n_components, decomp_params) + [seccol_params]
        return self.separate_params

    def split_as_unified_params(self, params, **kwargs):
        class NotSupportedError(Exception): pass
        raise NotSupportedError()

    def make_bounds_mask(self):
        bounds_mask = np.zeros(self.num_params + NUM_COL_PARAMS, dtype=bool)
        n = self.n_components
        nc = n - 1
        bounds_mask[0:nc*5] = True              # xr_params
        sep = nc + self.num_baseparams

        if self.integral_baseline:
            bounds_mask[sep-1] = True           # xr baseline fouling
            bounds_mask[self.pos[6]-1] = True   # uv aseline fouling

        bounds_mask[sep:sep+nc] = True      # rgs

        return bounds_mask

    def get_rg_start_index(self):
        return self.pos[2]

    def get_mr_start_index(self):
        return self.pos[6]

    def get_parameter_names(self):
        xr_names, rg_names, mapping_names, uv_names, mr_names, seccol_names = get_common_parameter_names(self.n_components - 1)
        xr_basenames = ["$xb_a$", "$xb_b$"]
        if self.num_baseparams == 3:
            xr_basenames += ["$xb_r$"]
        uv_basenames = ["$L$", "$x_0$", "$k$", "$b$", "$s_1$", "$s_2$", "$diffratio$"]
        if self.num_baseparams == 3:
            uv_basenames += ["$ub_r$"]
        return np.array(xr_names + xr_basenames + rg_names + mapping_names + uv_names + uv_basenames + mr_names + seccol_names)

    def get_trans_indeces(self):
        # never called
        assert False

    def split_get_unified_sec_params(self, params):
        # should not be called
        assert False

    def compute_comformance(self, xr_params, rg_params, edm_colparams, **kwargs):
        # this is made in accordance to .SimpleSecParams.sec_comformance

        from molass_legacy.SecTheory.ColumnConstants import SECCONF_LOWER_BOUND, BAD_CONFORMANCE_REDUCE

        stdev = np.std(np.concatenate([xr_params[:,0], edm_colparams]))     # edm_colparams == [Tz]
        log_conformance = np.log10(stdev)

        if log_conformance > 0:
            log_conformance *= BAD_CONFORMANCE_REDUCE   # large conformance at early stages can be misleading

        return max(SECCONF_LOWER_BOUND, log_conformance)

    def get_paramslider_info(self, devel=True):
        if devel:
            from importlib import reload
            import molass_legacy.ModelParams.EdmSliderInfo
            reload(molass_legacy.ModelParams.EdmSliderInfo)
        from .EdmSliderInfo import EdmSliderInfo
        nc = self.n_components - 1
        return EdmSliderInfo(nc=nc)


"""
    UV.LrfMappingDemo.py

    Copyright (c) 2023, SAXS Team, KEK-PF
"""
import numpy as np
from bisect import bisect_right
import molass_legacy.KekLib.DebugPlot as plt
from  MatrixData import simple_plot_3d
import MatplotlibUtils

def demo(caller):
    print("demo")
    dsets = caller.optinit_info.dsets
    optimizer = caller.fullopt
    params = caller.canvas.get_current_params()
    split_params = optimizer.split_params_simple(params)
    lrf_info = optimizer.objective_func(params, return_lrf_info=True, lrf_debug=False)
    Pxr, Cxr, Puv, Cuv, mapped_UvD = lrf_info.matrices

    uvD = optimizer.uvD
    wv = optimizer.wvector

    assert Cxr.shape[1] == Cuv.shape[1]
    elen = Cxr.shape[1]
    a, b = split_params[3]
    jxr = np.arange(elen) + optimizer.xr_curve.x[0]
    juv = a*jxr + b

    for M in (Pxr, Cxr, Puv, Cuv, mapped_UvD):
        print(M.shape)

    i280 = bisect_right(wv, 280)

    def plot_components(title, ax, ecurve, j_, P, C):
        ax.set_title(title)
        ax.plot(ecurve.x, ecurve.y)
        cy_list = []
        for k, c in enumerate(C):
            cy = c*P[i280,k]
            cy_list.append(cy)
            ax.plot(j_, cy, ":")
        ty = np.sum(cy_list, axis=0)
        ax.plot(j_, ty, ":", color="red")

    def plot_foldedness(title, ax, wv, P):
        ax.set_title(title)
        fy_list = []
        for k, p in enumerate(P.T[0:-1]):
            f = foldedness.compute(p)
            ax.plot(wv, p, ":", label="component-%d, Foldedness=%.1f" % (k+1, f))
        ax.legend()

    uv_curve = optimizer.uv_curve
    xr_curve = optimizer.xr_curve

    wv = optimizer.wvector
    xr_x = xr_curve.x
    uv_x = a*xr_x + b

    with plt.Dp():
        fig = plt.figure(figsize=(12,10))
        ax00 = fig.add_subplot(221)
        ax01 = fig.add_subplot(222)
        ax10 = fig.add_subplot(223, projection="3d")
        ax11 = fig.add_subplot(224, projection="3d")
        fig.suptitle("LRF Mapping Demo", fontsize=20)
        plot_components("UV Decomposition", ax00, uv_curve, juv, Puv, Cuv)
        plot_components("XR Decomposition", ax01, xr_curve, jxr, Pxr, Cxr)
        ax10.set_title("UV 3D View")
        simple_plot_3d(ax10, mapped_UvD, x=wv, y=uv_x)
        ax11.set_title("XR 3D View")
        simple_plot_3d(ax11, optimizer.xrD, x=optimizer.qvector, y=xr_x)
        fig.tight_layout()
        plt.show()

    with plt.Dp():
        fig, ax = plt.subplots(subplot_kw=dict(projection="3d"), figsize=(6,5))
        ax.set_title("UV 3D View with a badly-mapped Peak")
        ax.set_box_aspect(aspect=(1, 4, 1))
        simple_plot_3d(ax, mapped_UvD, x=wv, y=uv_x)
        simple_plot_3d(ax, mapped_UvD * 0.1, x=wv, color="red", alpha=0.1)

        ax.annotate3D('Bug',(300, 100, 0.0),
                      xytext=(40,30),
                      textcoords='offset points',
                      bbox=dict(boxstyle="round", fc="lightyellow"),
                      arrowprops = dict(arrowstyle="-|>",ec='black', fc='white', lw=3))

        fig.tight_layout()
        plt.show()
